"""
The model and the graph builder.
"""

from __future__ import annotations

import logging
import re
from collections import Counter
from collections.abc import Iterable
from copy import deepcopy
from types import MappingProxyType
from typing import IO, Any, TypeVar

import dill
import jax
import jax.random
import networkx as nx
import tensorflow_probability.substrates.jax.bijectors as jb
import tensorflow_probability.substrates.jax.distributions as jd
import tensorflow_probability.substrates.numpy.bijectors as nb
import tensorflow_probability.substrates.numpy.distributions as nd

from .nodes import (
    ArgGroup,
    Array,
    Bijector,
    Calc,
    Data,
    Dist,
    Group,
    InputGroup,
    Node,
    NodeState,
    TransientIdentity,
    Var,
)
from .viz import plot_nodes, plot_vars

__all__ = ["GraphBuilder", "Model", "load_model", "save_model"]

logger = logging.getLogger(__name__)


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Graph builder ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


NV = TypeVar("NV", Node, Var)


def _reduced_sum(*args: Array) -> Array:
    """Computes the sum after reducing arrays to scalars."""
    reduced = (arg.sum() if hasattr(arg, "sum") else arg for arg in args)
    return sum(reduced)


def _transform_back(var_transformed: Var) -> Calc:
    """
    Creates a :class:`.Calc` mapping a transformed parameter back to
    the original domain.
    """

    if var_transformed.dist_node is None:
        raise RuntimeError(
            f"{repr(var_transformed)} must have a transformed distribution"
        )

    transformed_distribution = var_transformed.dist_node.distribution

    def fn(at, *args, **kwargs):
        bijector = transformed_distribution(*args, **kwargs).bijector
        return bijector.inverse(at)

    inputs = var_transformed.dist_node.inputs
    kwinputs = var_transformed.dist_node.kwinputs

    return Calc(fn, var_transformed.value_node, *inputs, **kwinputs)  # type: ignore


class GraphBuilder:
    """
    A graph builder to prepare a :class:`.Model`.

    Constructs a model containing all nodes and variables that were added to the graph
    builder and their recursive inputs. The outputs of the nodes are not added to the
    model automatically, so the root nodes always need to be added explicitly.

    The standard workflow is to create the nodes and variables, add them to a graph
    builder, and construct a model from the graph builder. After the model has been
    constructed, some methods of the graph builder are not available anymore.

    Example
    -------
    >>> a = lsl.Var(1.0, name="a")
    >>> b = lsl.Var(2.0, name="b")
    >>> c = lsl.Var(lsl.Calc(lambda x, y: x + y, a, b), name="c")
    >>> gb = lsl.GraphBuilder()
    >>> gb.add(c)
    GraphBuilder(0 nodes, 1 vars)
    >>> model = gb.build_model()
    >>> model
    Model(9 nodes, 3 vars)
    >>> c.value
    3.0
    >>> gb.vars
    []
    """

    def __init__(self):
        self.nodes: list[Node] = []
        """The nodes that were explicitly added to the graph."""

        self.vars: list[Var] = []
        """The variables that were explicitly added to the graph."""

        self._log_lik_node: Node | None = None
        self._log_prior_node: Node | None = None
        self._log_prob_node: Node | None = None

    def _add_model_log_lik_node(self) -> GraphBuilder:
        """Adds the model log-likelihood node with the name ``_model_log_lik``."""

        if self.log_lik_node:
            self.add(TransientIdentity(self.log_lik_node, _name="_model_log_lik"))
            return self

        _, _vars = self._all_nodes_and_vars()
        inputs = (v.dist_node for v in _vars if v.has_dist and v.observed)
        node = Calc(_reduced_sum, *inputs, _name="_model_log_lik")
        self.add(node)
        return self

    def _add_model_log_prior_node(self) -> GraphBuilder:
        """Adds the model log-prior node with the name ``_model_log_prior``."""

        if self.log_prior_node:
            self.add(TransientIdentity(self.log_prior_node, _name="_model_log_prior"))
            return self

        _, _vars = self._all_nodes_and_vars()
        inputs = (v.dist_node for v in _vars if v.has_dist and v.parameter)
        node = Calc(_reduced_sum, *inputs, _name="_model_log_prior")
        self.add(node)
        return self

    def _add_model_log_prob_node(self) -> GraphBuilder:
        """Adds the model log-probability node with the name ``_model_log_prob``."""

        if self.log_prob_node:
            self.add(TransientIdentity(self.log_prob_node, _name="_model_log_prob"))
            return self

        nodes, _ = self._all_nodes_and_vars()
        inputs = (n for n in nodes if isinstance(n, Dist))
        node = Calc(_reduced_sum, *inputs, _name="_model_log_prob")
        self.add(node)
        return self

    def _add_model_seed_nodes(self) -> GraphBuilder:
        """Adds the model seed nodes with the names ``_model_*_seed``."""
        nodes, _ = self._all_nodes_and_vars()

        for node in nodes:
            if node.needs_seed:
                seed = Data(jax.random.PRNGKey(0), _name=f"_model_{node.name}_seed")
                node.set_inputs(*node.inputs, **{"seed": seed} | node.kwinputs)

        return self

    def _all_nodes_and_vars(self) -> tuple[list[Node], list[Var]]:
        """
        Returns all nodes and variables that were explicitly or implicitly
        (as recursive inputs) added to the graph.
        """
        nodes = self.nodes.copy()
        nodes.extend(node for var in self.vars for node in var.nodes)
        nodes = list(dict.fromkeys(nodes))

        if self.log_lik_node:
            nodes.append(self.log_lik_node)

        if self.log_prior_node:
            nodes.append(self.log_prior_node)

        if self.log_prob_node:
            nodes.append(self.log_prob_node)

        all_nodes: list[Node] = []
        all_vars: list[Var] = []

        while nodes:
            node = nodes.pop()

            if node in all_nodes:
                continue

            nodes.extend(node.all_input_nodes())
            all_nodes.append(node)

            if node.var:
                if node.var in all_vars:
                    continue

                nodes.extend(node.var.nodes)
                all_vars.append(node.var)

        return all_nodes, all_vars

    @staticmethod
    def _do_set_missing_names(nodes_or_vars: Iterable[NV], prefix: str) -> None:
        """Sets the missing names for the given nodes or variables."""
        other = [nv.name for nv in nodes_or_vars if nv.name]
        counter = -1

        for nv in nodes_or_vars:
            if not nv.name:
                name = f"{prefix}{(counter := counter + 1)}"

                while name in other:
                    name = f"{prefix}{(counter := counter + 1)}"

                nv.name = name
                other.append(name)

    def _set_missing_names(self) -> GraphBuilder:
        """Sets the missing node and variable names."""
        nodes, _vars = self._all_nodes_and_vars()
        self._do_set_missing_names(_vars, prefix="v")
        self._do_set_missing_names(nodes, prefix="n")
        return self

    def add(
        self, *args: Node | Var | GraphBuilder, to_float32: bool = True
    ) -> GraphBuilder:
        """
        Adds nodes, variables or other graph builders to the graph.

        Parameters
        ----------
        to_float32
            Whether to convert the dtype of the values of the added nodes \
            from float64 to float32.
        """

        for arg in args:
            if isinstance(arg, Node):
                self.nodes.append(arg)
            elif isinstance(arg, Var):
                self.vars.append(arg)
            elif isinstance(arg, GraphBuilder):
                self.nodes.extend(arg.nodes)
                self.vars.extend(arg.vars)
            else:
                raise RuntimeError(f"Cannot add {type(arg).__name__} to graph builder")

        if to_float32:
            self.convert_dtype("float64", "float32")

        return self

    def add_groups(self, *groups: Group, to_float32: bool = True) -> GraphBuilder:
        """
        Adds groups to the graph.

        Parameters
        ----------
        to_float32
            Whether to convert the dtype of the values of the added nodes \
            from float64 to float32.
        """

        for group in groups:
            old = self.groups()

            if group.name in old and group is not old[group.name]:
                raise RuntimeError(
                    f"Group with name {repr(group.name)} already exists "
                    "in graph builder"
                )

            self.add(*group.nodes_and_vars.values())

        if to_float32:
            self.convert_dtype("float64", "float32")

        return self

    def build_model(self, copy: bool = False) -> Model:
        """
        Builds a model from the graph.

        Constructs a model containing all nodes and variables that were added to the
        graph builder and their recursive inputs. The outputs of the nodes are not
        added to the model automatically, so the root nodes always need to be added
        explicitly.

        The standard workflow is to create the nodes and variables, add them to a graph
        builder, and construct a model from the graph builder. After the model has been
        constructed, some methods of the graph builder are not available anymore.

        Notes
        -----
        If this method is called with the argument ``copy=False``, all nodes and
        variables are removed from the graph builder, because most methods of the
        graph builder do not work with nodes that are part of a model.

        Example
        -------
        >>> a = lsl.Var(1.0, name="a")
        >>> b = lsl.Var(2.0, name="b")
        >>> c = lsl.Var(lsl.Calc(lambda x, y: x + y, a, b), name="c")
        >>> gb = lsl.GraphBuilder()
        >>> gb.add(c)
        GraphBuilder(0 nodes, 1 vars)
        >>> model = gb.build_model()
        >>> model
        Model(9 nodes, 3 vars)
        >>> c.value
        3.0
        >>> gb.vars
        []

        Parameters
        ----------
        copy
            Whether the nodes and variables should be copied when building the model.
        """
        nodes, _vars = self._all_nodes_and_vars()

        if not nodes:
            logger.warning("No nodes in graph builder, building an empty model")

        for node in nodes:
            if node.name.startswith("_model"):
                raise RuntimeError(f"{repr(node)} has reserved name '_model*'")

        gb = self.copy()
        gb._set_missing_names()
        gb._add_model_log_lik_node()
        gb._add_model_log_prior_node()
        gb._add_model_log_prob_node()
        gb._add_model_seed_nodes()

        nodes, _vars = gb._all_nodes_and_vars()
        nodes_and_vars = nodes + _vars

        model = Model(nodes_and_vars, grow=False, copy=copy)

        if not copy:
            self.nodes.clear()
            self.vars.clear()

            self._log_lik_node = None
            self._log_prior_node = None
            self._log_prob_node = None

        return model

    def convert_dtype(
        self, from_dtype: str | jax.numpy.dtype, to_dtype: str | jax.numpy.dtype
    ) -> GraphBuilder:
        """
        Tries to convert the node values in the graph to the specified data type.

        Works for nodes whose value is an array or pytree_. Nodes whose value is of
        another type are silently ignored.

        .. _pytree: https://jax.readthedocs.io/en/latest/pytrees.html
        """
        nodes, _ = self._all_nodes_and_vars()

        class ConversionWrapper:
            def __init__(self, value):
                self.value = value
                self.converted = False

                try:
                    if value.dtype == from_dtype:
                        self.value = value.astype(to_dtype)
                        self.converted = True
                except AttributeError:
                    pass

        for node in nodes:
            try:
                wrappers = jax.tree_map(ConversionWrapper, node.value)

                value = jax.tree_map(lambda x: x.value, wrappers)
                node.value = value  # type: ignore # data node

                converted = jax.tree_map(lambda x: x.converted, wrappers)

                if any(jax.tree_util.tree_flatten(converted)[0]):
                    logger.info(f"Converted dtype of {repr(node)}.value")
            except AttributeError:
                pass

        return self

    def copy(self) -> GraphBuilder:
        """Returns a shallow copy of the graph builder."""
        gb = GraphBuilder()
        gb.nodes = self.nodes.copy()
        gb.vars = self.vars.copy()

        gb.log_lik_node = self.log_lik_node
        gb.log_prior_node = self.log_prior_node
        gb.log_prob_node = self.log_prob_node

        return gb

    def count_node_names(self) -> dict[str, int]:
        """Counts the number of times each node name occurs in the graph."""
        nodes, _ = self._all_nodes_and_vars()
        counter = Counter(node.name for node in nodes if node.name)
        return dict(counter.most_common())

    def count_var_names(self) -> dict[str, int]:
        """Counts the number of times each variable name occurs in the graph."""
        _, _vars = self._all_nodes_and_vars()
        counter = Counter(var.name for var in _vars if var.name)
        return dict(counter.most_common())

    def groups(self) -> dict[str, Group]:
        """Collects the groups from all nodes and variables."""
        nodes, _vars = self._all_nodes_and_vars()
        g1 = {g.name: g for n in nodes for g in n.groups.values()}
        g2 = {g.name: g for v in _vars for g in v.groups.values()}
        return g1 | g2

    @property
    def log_lik_node(self) -> Node | None:
        """The user-defined log-likelihood node."""
        return self._log_lik_node

    @log_lik_node.setter
    def log_lik_node(self, log_lik_node: Node | None):
        if log_lik_node and not isinstance(log_lik_node, Node):
            raise RuntimeError("The log-likelihood node must be a node, not var")

        self._log_lik_node = log_lik_node

    @property
    def log_prior_node(self) -> Node | None:
        """The user-defined log-prior node."""
        return self._log_prior_node

    @log_prior_node.setter
    def log_prior_node(self, log_prior_node: Node | None):
        if log_prior_node and not isinstance(log_prior_node, Node):
            raise RuntimeError("The log-prior node must be a node, not var")

        self._log_prior_node = log_prior_node

    @property
    def log_prob_node(self) -> Node | None:
        """The user-defined log-probability node."""
        return self._log_prob_node

    @log_prob_node.setter
    def log_prob_node(self, log_prob_node: Node | None):
        if log_prob_node and not isinstance(log_prob_node, Node):
            raise RuntimeError("The log-probability node must be a node, not var")

        self._log_prob_node = log_prob_node

    def plot_nodes(self) -> GraphBuilder:
        """Plots all nodes in the graph."""
        nodes, _vars = self._all_nodes_and_vars()
        nodes_and_vars = nodes + _vars

        self._set_missing_names()
        model = Model(nodes_and_vars, grow=False)
        plot_nodes(model)

        model.pop_nodes_and_vars()

        return self

    def plot_vars(self) -> GraphBuilder:
        """Plots all variables in the graph."""
        nodes, _vars = self._all_nodes_and_vars()
        nodes_and_vars = nodes + _vars

        self._set_missing_names()
        model = Model(nodes_and_vars, grow=False)
        plot_vars(model)

        model.pop_nodes_and_vars()

        return self

    def rename(self, pattern: str, replacement: str) -> GraphBuilder:
        """Renames all nodes and variables in the graph."""
        self.rename_nodes(pattern, replacement)
        self.rename_vars(pattern, replacement)
        return self

    def rename_nodes(self, pattern: str, replacement: str) -> GraphBuilder:
        """Renames all nodes in the graph."""
        nodes, _ = self._all_nodes_and_vars()

        for node in nodes:
            if node.name:
                node.name = re.sub(pattern, replacement, node.name)

        return self

    def rename_vars(self, pattern: str, replacement: str) -> GraphBuilder:
        """Renames all variables in the graph."""
        _, _vars = self._all_nodes_and_vars()

        for var in _vars:
            if var.name:
                var.name = re.sub(pattern, replacement, var.name)

        return self

    def replace_node(self, old: Node, new: Node) -> GraphBuilder:
        """Replaces the ``old`` with the ``new`` node."""
        self.nodes = [new if x is old else x for x in self.nodes]
        nodes, _ = self._all_nodes_and_vars()

        for node in nodes:
            inputs = [new if x is old else x for x in node.inputs]
            kwinputs = {k: new if v is old else v for k, v in node.kwinputs.items()}
            node.set_inputs(*inputs, **kwinputs)

        return self

    def replace_var(self, old: Var, new: Var) -> GraphBuilder:
        """Replaces the ``old`` with the ``new`` variable."""
        self.vars = [new if x is old else x for x in self.vars]
        self.replace_node(old.var_value_node, new.var_value_node)
        self.replace_node(old.value_node, new.value_node)

        if old.dist_node:
            if not new.dist_node:
                raise RuntimeError(
                    f"Cannot replace {repr(old)} with distribution "
                    f"with {repr(new)} without distribution"
                )

            self.replace_node(old.dist_node, new.dist_node)

        return self

    def transform(
        self, var: Var, bijector: type[Bijector] | None = None, *args, **kwargs
    ) -> Var:
        """
        Transforms a variable by adding a new transformed variable as an input.

        Creates a new variable on the unconstrained space ``R**n`` with the appropriate
        transformed distribution, turning the original variable into a weak variable
        without an associated distribution. The transformation is performed using
        TFP's bijector classes.

        The value of the attribute :attr:`~liesel.model.nodes.Var.parameter` is
        transferred to the transformed variable and set to ``False`` on the original
        variable. The attributes :attr:`~liesel.model.nodes.Var.observed` and
        :attr:`~liesel.model.nodes.Var.role` are set to the default values for
        the transformed variable and remain unchanged on the original variable.

        Parameters
        ----------
        var
            The variable to transform (and add to the graph).
        bijector
            The bijector used to map the new transformed variable to this variable \
            (forward transformation). If ``None``, the experimental default event \
            space bijector (see TFP documentation) is used.
        args
            The arguments passed on to the init function of the bijector.
        kwargs
            The keyword arguments passed on to the init function of the bijector.

        Returns
        -------
        The new transformed variable which acts as an input to this variable.

        Raises
        ------
        RuntimeError
            If the variable is weak, has no TFP distribution, the distribution has
            no default event space bijector and the argument ``bijector`` is ``None``,
            or the local model for the variable cannot be built.
        """

        if var.weak:
            raise RuntimeError(f"{repr(var)} is weak")

        if var.dist_node is None:
            raise RuntimeError(f"{repr(var)} has no distribution")

        # avoid name clashes
        self._set_missing_names()

        try:
            Model([var])
        except Exception:
            raise RuntimeError(f"Cannot build local model for {repr(var)}")

        self.add(var)

        # if we got this far, we can assume:
        # - the var and its inputs have numeric values
        # - the var and its inputs are up-to-date

        tfp_dist = var.dist_node.init_dist()
        default_bijector = tfp_dist.experimental_default_event_space_bijector()
        has_default_bijector = default_bijector is not None
        use_default_bijector = bijector is None

        if use_default_bijector and not has_default_bijector:
            raise RuntimeError(
                f"{repr(var)} has distribution without default event space bijector "
                "and no bijector was given"
            )

        if isinstance(tfp_dist, jd.Distribution):
            tfd = jd
            tfb = jb
        elif isinstance(tfp_dist, nd.Distribution):
            tfd = nd
            tfb = nb
        else:
            raise RuntimeError(f"{repr(var)} has no TFP distribution")

        # no copy necessary:
        # >>> from copy import copy
        # >>> import tensorflow_probability.substrates.numpy.distributions as tfd
        # >>> CopiedNormal = copy(tfd.Normal)
        # >>> CopiedNormal is tfd.Normal
        # True

        tfp_dist_cls = var.dist_node.distribution

        dist_inputs = InputGroup(
            *var.dist_node.inputs,
            **var.dist_node.kwinputs,  # type: ignore
        )

        bijector_inputs = InputGroup(*args, **kwargs)

        # define distribution "class" for the transformed var
        def make_transformed_distribution(dist_args: ArgGroup, bijector_args: ArgGroup):
            tfp_dist = tfp_dist_cls(*dist_args.args, **dist_args.kwargs)

            if bijector is None:
                bijector_obj = tfp_dist.experimental_default_event_space_bijector(
                    *bijector_args.args, **bijector_args.kwargs
                )

                bijector_inv = tfb.Invert(bijector_obj)
            else:
                bijector_obj = bijector(*bijector_args.args, **bijector_args.kwargs)
                bijector_inv = tfb.Invert(bijector_obj)

            return tfd.TransformedDistribution(
                tfp_dist, bijector_inv, validate_args=tfp_dist.validate_args
            )

        # build transformed var
        dist_node_transformed = Dist(
            make_transformed_distribution, dist_inputs, bijector_inputs
        )

        # transfer flags
        dist_node_transformed.needs_seed = var.dist_node.needs_seed
        dist_node_transformed.per_obs = var.dist_node.per_obs

        # transform value
        bijector_obj = dist_node_transformed.init_dist().bijector
        value_transformed = bijector_obj.forward(var.value)

        name_transformed = f"{var.name}_transformed" if var.name else ""

        var_transformed = Var(
            value_transformed, dist_node_transformed, name_transformed
        )

        var_transformed.parameter = var.parameter

        # var is now the forward transformation (a weak node without distribution)
        var.value_node = _transform_back(var_transformed)
        var.dist_node = None
        var.parameter = False

        return var_transformed

    def update(self) -> GraphBuilder:
        """Updates all nodes in the graph."""
        nodes, _vars = self._all_nodes_and_vars()
        nodes_and_vars = nodes + _vars

        self._set_missing_names()
        model = Model(nodes_and_vars, grow=False)
        model.pop_nodes_and_vars()

        return self

    def __repr__(self) -> str:
        brackets = f"({len(self.nodes)} nodes, {len(self.vars)} vars)"
        return type(self).__name__ + brackets


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Model ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


class Model:
    """
    A model with a static graph.

    Parameters
    ----------
    nodes_and_vars
        The nodes and variables to include in the model.
    grow
        Whether a :class:`.GraphBuilder` should be used to grow the model (finding \
        the recursive inputs of the nodes and variables), and to add the model nodes.
    copy
        Whether the nodes and variables should be copied upon initialization.
    """

    def __init__(
        self,
        nodes_and_vars: Iterable[Node | Var],
        grow: bool = True,
        copy: bool = False,
    ):
        if grow:
            model = GraphBuilder().add(*nodes_and_vars).build_model()
            nodes_and_vars = [*model.nodes.values(), *model.vars.values()]
            model.pop_nodes_and_vars()

        nodes = [nv for nv in nodes_and_vars if isinstance(nv, Node)]
        nodes = list(dict.fromkeys(nodes).keys())
        counts = Counter(n.name for n in nodes)
        dups = [k for k, v in counts.items() if v > 1]

        if dups:
            raise RuntimeError(f"Duplicate node names: {', '.join(dups)}")

        _vars = [nv for nv in nodes_and_vars if isinstance(nv, Var)]
        _vars = list(dict.fromkeys(_vars).keys())
        counts = Counter(v.name for v in _vars)
        dups = [k for k, v in counts.items() if v > 1]

        if dups:
            raise RuntimeError(f"Duplicate variable names: {', '.join(dups)}")

        groups = [g for nv in nodes_and_vars for g in nv.groups.values()]
        groups = list(dict.fromkeys(groups).keys())
        counts = Counter(g.name for g in groups)
        dups = [k for k, v in counts.items() if v > 1]

        if dups:
            raise RuntimeError(f"Duplicate group names: {', '.join(dups)}")

        self._nodes = {n.name: n for n in nodes}
        self._vars = {v.name: v for v in _vars}

        if copy:
            self._nodes, self._vars = deepcopy((self._nodes, self._vars))

        for node in self._nodes.values():
            node._clear_outputs()
            node._set_model(self)

        for node in self._nodes.values():
            for _input in node.all_input_nodes():
                _input._add_output(node)

        self._node_graph = self._build_node_graph(self._nodes.values())
        self._var_graph = self._build_var_graph(self._vars.values())

        self._sorted_nodes = list(nx.topological_sort(self._node_graph))

        self._auto_update = True
        self._seed_nodes = []

        for node in self._sorted_nodes:
            if node.name.startswith("_model_") and node.name.endswith("_seed"):
                self._seed_nodes.append(node)

            node.update()

    @staticmethod
    def _build_node_graph(nodes: Iterable[Node]) -> nx.DiGraph:
        """Builds the directed graph of the model nodes."""
        edges: list[tuple[Node, Node]] = []

        for node in nodes:
            edges.extend((_input, node) for _input in node.all_input_nodes())

        graph = nx.DiGraph(edges)
        graph.add_nodes_from(nodes)
        return graph

    @staticmethod
    def _build_var_graph(_vars: Iterable[Var]) -> nx.DiGraph:
        """Builds the directed graph of the model variables."""
        edges: list[tuple[Var, Var]] = []

        for var in _vars:
            edges.extend((_input, var) for _input in var.all_input_vars())

        graph = nx.DiGraph(edges)
        graph.add_nodes_from(_vars)
        return graph

    def _copy_computational_model(self) -> Model:
        """Returns a deep copy of the model with all node states cleared."""
        backup = self.state

        for node in self._nodes.values():
            node.clear_state()

        empty = deepcopy(self)
        self.state = backup

        return empty

    def _recursive_inputs(self, name: str) -> list[Node]:
        """Returns the recursive inputs of a model node."""
        nodes = [self._nodes[name]]
        visited = []

        while nodes:
            node = nodes.pop()

            if node in visited:
                continue

            nodes.extend(node.all_input_nodes())
            visited.append(node)

        return visited

    @property
    def auto_update(self) -> bool:
        """
        Whether to update the model automatically if the value of a node is modified.

        The auto-update can be disabled to improve the performance if multiple model
        parameters are updated at once.
        """
        return self._auto_update

    @auto_update.setter
    def auto_update(self, auto_update: bool):
        self._auto_update = auto_update

    def groups(self) -> dict[str, Group]:
        """Collects the groups from all nodes and variables."""
        g1 = {g.name: g for n in self._nodes.values() for g in n.groups.values()}
        g2 = {g.name: g for v in self._vars.values() for g in v.groups.values()}
        return g1 | g2

    def copy_nodes_and_vars(self) -> tuple[dict[str, Node], dict[str, Var]]:
        """Returns an unfrozen deep copy of the model nodes and variables."""
        nodes, _vars = deepcopy((self._nodes, self._vars))

        for node in nodes.values():
            node._unset_model()

        nodes = {nm: nd for nm, nd in nodes.items() if not nm.startswith("_model")}

        return nodes, _vars

    @property
    def log_lik(self) -> Array:
        """
        The log-likelihood of the model.

        Defined as the sum of the log-probabilities of all observed variables
        with a probability distribution.
        """
        return self._nodes["_model_log_lik"].value

    @property
    def log_prior(self) -> Array:
        """
        The log-prior of the model.

        Defined as the sum of the log-probabilities of all parameter variables
        with a probability distribution.
        """
        return self._nodes["_model_log_prior"].value

    @property
    def log_prob(self) -> Array:
        """
        The (unnormalized) log-probability / log-posterior of the model.

        Defined as the sum of all distribution nodes.
        """
        return self._nodes["_model_log_prob"].value

    @property
    def node_graph(self) -> nx.DiGraph:
        """The directed graph of the model nodes."""
        return self._node_graph

    @property
    def nodes(self) -> MappingProxyType[str, Node]:
        """A mapping of the model nodes with their names as keys."""
        return MappingProxyType(self._nodes)

    def pop_nodes_and_vars(self) -> tuple[dict[str, Node], dict[str, Var]]:
        """
        Pops the nodes and variables out of this model.

        All nodes and variables are unfrozen and their reference to this model
        is removed. This model becomes invalid and cannot be used anymore.
        """
        nodes = self._nodes.copy()
        _vars = self._vars.copy()

        for node in nodes.values():
            node._unset_model()

        nodes = {nm: nd for nm, nd in nodes.items() if not nm.startswith("_model")}

        # clear the model
        self._nodes.clear()
        self._vars.clear()
        self._node_graph.clear()
        self._var_graph.clear()
        self._sorted_nodes.clear()
        self._seed_nodes.clear()

        return nodes, _vars

    def set_seed(self, seed: jax.random.KeyArray) -> Model:
        """
        Splits and sets the seed / PRNG key.

        Parameters
        ----------
        seed
            The seed is split and distributed to the seed nodes of the model.
            Must be a ``KeyArray``, i.e. an array of shape (2,) and dtype ``uint32``.
            See :mod:`jax.random` for more details.
        """
        seeds = jax.random.split(seed, len(self._seed_nodes))

        for node, seed in zip(self._seed_nodes, seeds):
            node.value = seed  # type: ignore  # data node

        return self

    @property
    def state(self) -> dict[str, NodeState]:
        """The state of the model as a dict of node names and states."""
        return {name: node.state for name, node in self._nodes.items()}

    @state.setter
    def state(self, state: dict[str, NodeState]):
        for name, node_state in state.items():
            self._nodes[name].state = node_state

    def update(self, *names: str) -> Model:
        """
        Updates the target nodes and their recursive inputs if they are outdated.

        The update is performed in a topological order, restoring a consistent state
        of the model. This method is called automatically by the nodes if their value
        is modified (unless :attr:`.auto_update` is ``False``).

        Parameters
        ----------
        names
            The names of the target nodes to be updated.
        """

        if not names:
            for node in self._sorted_nodes:
                if node.outdated:
                    node.update()
        else:
            inputs = set().union(*(self._recursive_inputs(name) for name in names))

            for node in self._sorted_nodes:
                if node in inputs and node.outdated:
                    node.update()

        return self

    @property
    def var_graph(self) -> nx.DiGraph:
        """The directed graph of the model variables."""
        return self._var_graph

    @property
    def vars(self) -> MappingProxyType[str, Var]:
        """A mapping of the model variables with their names as keys."""
        return MappingProxyType(self._vars)

    def __repr__(self) -> str:
        brackets = f"({len(self._nodes)} nodes, {len(self._vars)} vars)"
        return type(self).__name__ + brackets


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Save and load models ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


def save_model(model: Any, file: str | IO[bytes]) -> None:
    """
    Saves a model to a `dill <https://github.com/uqfoundation/dill>`_ file.

    Parameters
    ----------
    model
        The model to be saved.
    file
        The file handler or path to save the model to.
    """

    if isinstance(file, str):
        with open(file, "wb") as handle:
            dill.dump(model, handle)
    else:
        dill.dump(model, file)


def load_model(file: str | IO[bytes]) -> Any:
    """
    Loads a model from a `dill <https://github.com/uqfoundation/dill>`_ file.

    Parameters
    ----------
    file
        The file handler or path to load the model from.
    """

    if isinstance(file, str):
        with open(file, "rb") as handle:
            model = dill.load(handle)
    else:
        model = dill.load(file)

    return model



"""
Posterior statistics and diagnostics.
"""

from __future__ import annotations

from collections.abc import Sequence
from typing import Any, NamedTuple

import arviz as az
import jax
import jax.numpy as jnp
import numpy as np
import pandas as pd
from deprecated.sphinx import deprecated

from liesel.goose.engine import ErrorLog, SamplingResults
from liesel.goose.pytree import slice_leaves, stack_leaves
from liesel.goose.types import Position
from liesel.option import Option


class ErrorSummaryForOneCode(NamedTuple):
    error_code: int
    error_msg: str
    count_per_chain: np.ndarray
    count_per_chain_posterior: None


ErrorSummary = dict[str, dict[int, ErrorSummaryForOneCode]]
"""
See docstring of ``_make_error_summary``.
"""


def _make_error_summary(
    error_log: ErrorLog,
    posterior_error_log: Option[ErrorLog],
) -> ErrorSummary:
    """
    Creates an error summary from the error log.

    The returned value looks like this::

        {
            kernel_identifier: {
                error_code: (error_code, error_msg, count, count_in_posterior),
                error_code: (error_code, error_msg, count, count_in_posterior),
                ...
            },
            ...
        }

    The ``error_msg`` is the empty string if the kernel class is not supplied in the
    ``error_log``.
    """
    error_summary = {}
    for kel in error_log.values():
        counter_dict: dict[int, np.ndarray] = {}

        # calculate the overall counts
        ec_unique = np.unique(kel.error_codes)
        for ec in ec_unique:
            if ec == 0:
                continue
            occurences_per_chain = np.sum(kel.error_codes == ec, axis=1)
            counter_dict[ec] = occurences_per_chain

        krnl_summary: dict[int, ErrorSummaryForOneCode] = {}
        for key, count in counter_dict.items():
            ec = key
            # type ignore is ok since the type must implement the kernel protocol.
            error_msg = kel.kernel_cls.map_or(
                "", lambda krn_cls: krn_cls.error_book[ec]  # type: ignore
            )
            krnl_summary[ec] = ErrorSummaryForOneCode(ec, error_msg, count, None)

        # calculate the counts in the posterior
        if posterior_error_log.is_some():
            posterior_error_log_unwrapped = posterior_error_log.unwrap()
            kel_post = posterior_error_log_unwrapped[kel.kernel_ident]
            for ec in ec_unique:
                if ec == 0:
                    continue
                occurences_per_chain = np.sum(kel_post.error_codes == ec, axis=1)
                krnl_summary[ec] = krnl_summary[ec]._replace(
                    count_per_chain_posterior=occurences_per_chain
                )

        error_summary[kel.kernel_ident] = krnl_summary

    return error_summary


class Summary:
    """
    A summary object.

    Offers two main use cases:

    1. View an overall summary by printing a summary instance, including a summary table
       of the posterior samples and a summary of sammpling errors.
    2. Programmatically access summary statistics via
       ``quantities[quantity_name][var_name]``. Please refer to the documentation of the
       attribute :attr:`.quantities` for details.

    Additionally, the summary object can be turned into a :class:`~pandas.DataFrame`
    using :meth:`.to_dataframe`.

    Parameters
    ----------
    results
        The sampling results to summarize.
    additional_chain
        can be supplied to add more parameters to the summary output. Must be a position
        chain which matches chain and time dimension of the posterior chain as returned
        by :meth:`.SamplingResults.get_posterior_samples`.
    hdi_prob
        Level on which to return posterior highest density intervals.
    selected, deselected
        Allow to get a summary only for a subset of the position keys.
    per_chain
        If *True*, the summary is calculated on a per-chain basis. Certain measures like
        ``rhat`` are not available if ``per_chain`` is *True*.

    Notes
    -----
    This class is still considered experimental. The API may still undergo larger
    changes.
    """

    per_chain: bool
    """
    Whether results are summarized for individual chains (*True*), or aggregated
    over chains (*False*).
    """
    quantities: dict[str, dict[str, np.ndarray]]
    """
    Dict of summarizing quantities.

    Built up in hierarchies as. Let ``summary`` be a :class:`.Summary` instance. The
    hierarchy is::

        q = summary.quantities["quantity_name"]["parameter_name"]

    The extracted object is an ``np.ndarray``. If ``per_chain=True``, the arrays for
    the ``"quantile"`` and ``"hdi"`` quantities have the following dimensions:

    1. First index refers to the chain
    2. Second index refers to the quantile/interval
    3. Third and subsequent indices refer to individual parameters.

    If ``per_chain=True``, the arrays for the other quantiles have the dimensions:

    1. First index refers to the chain
    2. Second and subsequent indices refer to individual parameters.

    If ``per_chain=False``, the first index is removed for all quantities.
    """
    config: dict
    """
    A dictionary of config settings for this summary object. Should NOT be changed
    after initialization; such changes have no effect on the computed summary values.
    """
    sample_info: dict
    """
    Dictionary of meta-information about the mcmc samples used to create this summary
    object.

    Contains ``num_chains``, ``sample_size_per_chain``, and ``warmup_size_per_chain``.
    """
    error_summary: ErrorSummary
    """
    Contains error information for each kernel.
    """
    kernels_by_pos_key: dict[str, str]
    """
    A dict, linking parameter names (the keys) to the kernel identifier (the values).
    The identifier refers to the kernel that was used to sample the respective
    parameter.
    """

    def __init__(
        self,
        results: SamplingResults,
        additional_chain: Position | None = None,
        quantiles: Sequence[float] = (0.05, 0.5, 0.95),
        hdi_prob: float = 0.9,
        selected: list[str] | None = None,
        deselected: list[str] | None = None,
        per_chain: bool = False,
    ):
        posterior_chain = results.get_posterior_samples()
        if additional_chain:
            for k, v in additional_chain.items():
                posterior_chain[k] = v

        if selected:
            posterior_chain = Position(
                {
                    key: value
                    for key, value in posterior_chain.items()
                    if key in selected
                }
            )

        if deselected is not None:
            for key in deselected:
                del posterior_chain[key]

        # get some general infos on the sampling
        param_chain = next(iter(posterior_chain.values()))
        epochs = results.positions.get_epochs()

        warmup_size = np.sum(
            [epoch.duration for epoch in epochs if epoch.type.is_warmup(epoch.type)]
        )

        sample_info = {
            "num_chains": param_chain.shape[0],
            "sample_size_per_chain": param_chain.shape[1],
            "warmup_size_per_chain": warmup_size,
        }

        # convert everything to numpy array
        for key in posterior_chain:
            posterior_chain[key] = np.asarray(posterior_chain[key])

        # calculate quantiles either per chain and merge the results or all at once
        single_chain_summaries = []
        if per_chain:
            for chain_idx in range(sample_info["num_chains"]):
                single_chain = slice_leaves(
                    posterior_chain, jnp.s_[None, chain_idx, ...]
                )

                qdict = _create_quantity_dict(single_chain, quantiles, hdi_prob)

                single_chain_summaries.append(qdict)

            quantities = stack_leaves(single_chain_summaries, axis=0)

        else:
            quantities = _create_quantity_dict(posterior_chain, quantiles, hdi_prob)

        config = {
            "quantiles": quantiles,
            "hdi_prob": hdi_prob,
            "chains_merged": not per_chain,
        }

        error_summary = _make_error_summary(
            results.get_error_log(False).unwrap(), results.get_error_log(True)
        )

        self.per_chain = per_chain
        self.quantities = quantities
        self.config = config
        self.sample_info = sample_info
        self.error_summary = error_summary
        self.kernels_by_pos_key = results.get_kernels_by_pos_key()

    def to_dataframe(self) -> pd.DataFrame:
        """Turns Summary object into a :class:`~pandas.DataFrame` object."""

        # don't change the original data
        quants = self.quantities.copy()

        # make new entries for the quantiles
        if self.per_chain:
            for i, q in enumerate(self.config["quantiles"]):
                quants[f"q_{q}"] = {
                    k: v[:, i, ...] for k, v in quants["quantile"].items()
                }

            quants["hdi_low"] = {k: v[:, 0, ...] for k, v in quants["hdi"].items()}
            quants["hdi_high"] = {k: v[:, 1, ...] for k, v in quants["hdi"].items()}
        else:
            for i, q in enumerate(self.config["quantiles"]):
                quants[f"q_{q}"] = {k: v[i, ...] for k, v in quants["quantile"].items()}

            quants["hdi_low"] = {k: v[0, ...] for k, v in quants["hdi"].items()}
            quants["hdi_high"] = {k: v[1, ...] for k, v in quants["hdi"].items()}

        # remove the old entries
        del quants["quantile"]
        del quants["hdi"]

        # create one row per entry
        df_dict = {}
        for var in quants["mean"].keys():
            it = np.nditer(quants["mean"][var], flags=["multi_index"])
            for _ in it:
                var_fqn = (
                    var if len(it.multi_index) == 0 else f"{var}{list(it.multi_index)}"
                )
                quant_per_elem: dict[str, Any] = {}
                quant_per_elem["variable"] = var
                quant_per_elem["kernel"] = self.kernels_by_pos_key.get(var, "-")
                if self.config["chains_merged"]:
                    quant_per_elem["var_index"] = it.multi_index
                    quant_per_elem["sample_size"] = (
                        self.sample_info["sample_size_per_chain"]
                        * self.sample_info["num_chains"]
                    )
                else:
                    quant_per_elem["chain_index"] = it.multi_index[0]
                    quant_per_elem["var_index"] = it.multi_index[1:]
                    quant_per_elem["sample_size"] = self.sample_info[
                        "sample_size_per_chain"
                    ]

                for quant_name, quant_dict in quants.items():
                    quant_per_elem[quant_name] = quant_dict[var][it.multi_index]

                # convert jax.Arrays (scalar) to floats so that pandas treats them
                # correctly
                for key, val in quant_per_elem.items():
                    if isinstance(val, jax.Array):
                        # value should be a scalar
                        assert val.shape == ()

                        # replace dict element with value casted to float32
                        quant_per_elem[key] = float(val)

                df_dict[var_fqn] = quant_per_elem

        # convert to dataframe and use varname as index
        df = pd.DataFrame.from_dict(df_dict, orient="index")
        df = df.reset_index()
        df = df.rename(columns={"index": "var_fqn"})
        df = df.set_index("variable")

        return df

    def _param_df(self):
        df = self.to_dataframe()

        df.index.name = "parameter"
        df = df.rename(columns={"var_index": "index"})
        df = df.set_index("index", append=True)

        qtls = [f"q_{qtl}" for qtl in self.config["quantiles"]]
        cols = (
            ["kernel", "mean", "sd"]
            + qtls
            + ["sample_size", "ess_bulk", "ess_tail", "rhat"]
        )
        cols = [col for col in cols if col in df.columns]
        df = df[cols]

        return df

    def _error_df(self, per_chain=False):
        # fmt: off
        df = pd.concat({
            kernel: pd.DataFrame.from_dict(code_summary, orient="index")
            for kernel, code_summary in self.error_summary.items()
        })
        # fmt: on

        if df.empty:
            return df

        df = df.reset_index(level=1, drop=True)
        df["error_code"] = df["error_code"].astype(int)
        df = df.set_index(["error_code", "error_msg"], append=True)
        df.index.names = ["kernel", "error_code", "error_msg"]

        # fmt: off
        df = df.rename(columns={
            "count_per_chain": "total",
            "count_per_chain_posterior": "posterior",
        })
        # fmt: on

        df = df.explode(["total", "posterior"])
        df["warmup"] = df["total"] - df["posterior"]
        df = df.drop(columns="total")

        df = df.melt(
            value_vars=["warmup", "posterior"],
            var_name="phase",
            value_name="count",
            ignore_index=False,
        )

        df["phase"] = pd.Categorical(df["phase"], categories=["warmup", "posterior"])

        df = df.set_index("phase", append=True)
        df["chain"] = df.groupby(level=[0, 1, 2, 3]).cumcount()
        df = df.set_index("chain", append=True)
        df = df.sort_index()

        df["sample_size"] = None
        warmup_size = self.sample_info["warmup_size_per_chain"]
        posterior_size = self.sample_info["sample_size_per_chain"]
        df.loc[pd.IndexSlice[:, :, :, "warmup"], "sample_size"] = warmup_size
        df.loc[pd.IndexSlice[:, :, :, "posterior"], "sample_size"] = posterior_size
        df["relative"] = df["count"] / df["sample_size"]
        df = df.drop(columns="sample_size")

        if not per_chain:
            df = df.groupby(level=[0, 1, 2, 3], observed=True)
            df = df.aggregate({"count": "sum", "relative": "mean"})
            df = df.sort_index()

        return df

    def __repr__(self):
        param_df = self._param_df()
        error_df = self._error_df()

        txt = "Parameter summary:\n\n" + repr(param_df)

        if not error_df.empty:
            txt += "\n\nError summary:\n\n" + repr(error_df)

        return txt

    def _repr_html_(self):
        param_df = self._param_df()
        error_df = self._error_df()

        html = "\n<p><strong>Parameter summary:</strong></p>\n" + param_df.to_html()

        if not error_df.empty:
            html += "\n<p><strong>Error summary:</strong></p>\n" + error_df.to_html()

        html += "\n"
        return html

    def _repr_markdown_(self):
        param_df = self._param_df()
        error_df = self._error_df()

        try:
            param_md = param_df.to_markdown()
            error_md = error_df.to_markdown()
        except ImportError:
            param_md = f"```\n{repr(param_df)}\n```"
            error_md = f"```\n{repr(error_df)}\n```"

        md = "\n\n**Parameter summary:**\n\n" + param_md

        if not error_df.empty:
            md += "\n\n**Error summary:**\n\n" + error_md

        md += "\n\n"
        return md

    def __str__(self):
        return str(self.to_dataframe())

    @classmethod
    @deprecated(reason="Functionality moved directly to the __init__.", version="0.1.4")
    def from_result(
        cls,
        result: SamplingResults,
        additional_chain: Position | None = None,
        quantiles: Sequence[float] = (0.05, 0.5, 0.95),
        hdi_prob: float = 0.9,
        selected: list[str] | None = None,
        deselected: list[str] | None = None,
        per_chain=False,
    ) -> Summary:
        """
        Alias for :meth:`.from_results` for backwards compatibility.

        In addition to the name, there are two further subtle differences to
        :meth:`.from_results`.

        - The argument ``result`` is in singular. The method :meth:`.from_results` uses
          the plural instead.
        - The argument ``result`` is of type :class:`.SamplingResult`, which itself is
          an alias for :meth:`.SamplingResults`.
        """

        return cls.from_results(
            results=result,
            additional_chain=additional_chain,
            quantiles=quantiles,
            hdi_prob=hdi_prob,
            selected=selected,
            deselected=deselected,
            per_chain=per_chain,
        )

    @staticmethod
    @deprecated(reason="Functionality moved directly to the __init__.", version="0.1.4")
    def from_results(
        results: SamplingResults,
        additional_chain: Position | None = None,
        quantiles: Sequence[float] = (0.05, 0.5, 0.95),
        hdi_prob: float = 0.9,
        selected: list[str] | None = None,
        deselected: list[str] | None = None,
        per_chain=False,
    ) -> Summary:
        """
        Creates a :class:`.Summary` object from a results object.
        """

        return Summary(
            results=results,
            additional_chain=additional_chain,
            quantiles=quantiles,
            hdi_prob=hdi_prob,
            selected=selected,
            deselected=deselected,
            per_chain=per_chain,
        )


def _create_quantity_dict(
    chain: Position, quantiles: Sequence[float], hdi_prob: float
) -> dict[str, dict[str, np.ndarray]]:
    azchain = az.convert_to_inference_data(chain).posterior

    # calculate quantities
    mean = azchain.mean(dim=["chain", "draw"])
    var = azchain.var(dim=["chain", "draw"])
    sd = azchain.std(dim=["chain", "draw"])
    quantile = azchain.quantile(q=quantiles, dim=["chain", "draw"])
    hdi = az.hdi(azchain, hdi_prob=hdi_prob)

    ess_bulk = az.ess(azchain, method="bulk")
    ess_tail = az.ess(azchain, method="tail")
    mcse_mean = az.mcse(azchain, method="mean")
    mcse_sd = az.mcse(azchain, method="sd")

    # place quantities in dict
    quantities = {
        "mean": mean,
        "var": var,
        "sd": sd,
        "quantile": quantile,
        "hdi": hdi,
        "rhat": None,
        "ess_bulk": ess_bulk,
        "ess_tail": ess_tail,
        "mcse_mean": mcse_mean,
        "mcse_sd": mcse_sd,
    }

    if azchain.chain.size > 1:
        quantities["rhat"] = az.rhat(azchain)
    else:
        del quantities["rhat"]

    # convert to simple dict[str, np.ndarray]
    for key, val in quantities.items():
        quantities[key] = {k: v.values for k, v in val.data_vars.items()}

    # hdi shape BEFORE
    # VarIDX --- HDI

    # special treatment for hdi since the function uses the last axis to refer
    # to the quantile
    for k, v in quantities["hdi"].items():
        quantities["hdi"][k] = np.moveaxis(v, -1, 0)

    # hdi shape AFTER
    # HDI --- VarIDX

    return quantities



from __future__ import annotations

from functools import partial

import jax
import numpy as np
import pytest

import liesel.goose as gs

try:
    import pymc as pm

    from liesel.experimental.pymc import PyMCInterface
except ImportError:
    pytestmark = pytest.mark.skip(reason="need pymc to run")

roughly_close = partial(np.allclose, rtol=0.1, atol=0.01)


@pytest.fixture
def basic_lm():
    RANDOM_SEED = 123
    rng = np.random.RandomState(RANDOM_SEED)

    # set parameter values
    num_obs = 100
    sigma = 1.0
    beta = [1.0, 1.0, 2.0]

    # simulate covariates
    x1 = rng.randn(num_obs)
    x2 = 0.5 * rng.randn(num_obs)

    # Simulate outcome variable
    y = beta[0] + beta[1] * x1 + beta[2] * x2 + sigma * rng.normal(size=num_obs)

    x1 = np.asarray(x1, np.float32)
    x2 = np.asarray(x2, np.float32)
    y = np.asarray(y, np.float32)

    basic_model = pm.Model()
    with basic_model:
        # priors
        beta = pm.Normal("beta", mu=0.0, sigma=10.0, shape=3)
        sigma = pm.HalfNormal(
            "sigma", sigma=1.0
        )  # automatically transformed to real via log

        # predicted value
        mu = beta[0] + beta[1] * x1 + beta[2] * x2

        # track the predicted value of the first obs
        pm.Deterministic("mu[0]", mu[0])

        # distribution of response (likelihood)
        pm.Normal("y_obs", mu=mu, sigma=sigma, observed=y)

    return basic_model


def test_simple():
    model = pm.Model()

    with model:
        _ = pm.Normal("mu", mu=0.0, sigma=1.0)

    interface = PyMCInterface(model=model)
    state = interface.get_initial_state()
    assert state["mu"] == 0.0
    assert roughly_close(interface.log_prob(state), -0.91893853)

    state = interface.update_state({"mu": np.array(1.0)}, state)
    assert state["mu"] == 1.0
    assert roughly_close(interface.log_prob(state), -1.41893853)


def test_default_dtype():
    model = pm.Model()
    with model:
        _ = pm.Normal("mu", mu=0.0, sigma=1.0)

    interface = PyMCInterface(model=model)
    state = interface.get_initial_state()

    assert not jax.config.read("jax_enable_x64")
    assert interface.log_prob(state).dtype == np.float32


def test_simple2():
    model = pm.Model()

    with model:
        sigma = pm.HalfNormal(
            "sigma",
            sigma=1.0,
        )
        _ = pm.Normal("mu", mu=0.0, sigma=sigma)

    interface = PyMCInterface(model=model, additional_vars=["sigma"])
    state = interface.get_initial_state()

    assert state["mu"] == 0.0
    assert state["sigma_log__"] == 0.0
    assert roughly_close(interface.log_prob(state), -0.91893853 + -0.72579135)
    assert interface.extract_position(["sigma"], state)["sigma"] == 1.0

    state = interface.update_state(
        {"mu": np.array(1.0), "sigma_log__": np.array(-1.0)}, state
    )
    assert state["mu"] == 1.0
    assert state["sigma_log__"] == -1.0
    assert roughly_close(interface.log_prob(state), -3.613467 + -1.29345899)


@pytest.mark.mcmc
def test_mcmc(basic_lm: pm.Model):  # type: ignore
    interface = PyMCInterface(basic_lm, additional_vars=["sigma", "mu[0]"])
    state = interface.get_initial_state()
    builder = gs.EngineBuilder(1, 2)
    builder.set_initial_values(state)
    builder.set_model(interface)
    builder.set_duration(1000, 2000)

    builder.add_kernel(gs.NUTSKernel(["beta", "sigma_log__"]))

    builder.positions_included = ["sigma", "mu[0]"]

    engine = builder.build()

    engine.sample_all_epochs()
    results = engine.get_results()
    sum = gs.Summary(results)

    assert roughly_close(
        sum.quantities["mean"]["beta"], [0.9068794, 0.94820887, 2.0448447]
    )
    assert roughly_close(sum.quantities["mean"]["sigma_log__"], -0.01403057)
    assert roughly_close(sum.quantities["mean"]["sigma"], 0.98853235)
    assert roughly_close(sum.quantities["mean"]["mu[0]"], 0.5339259)


@pytest.mark.mcmc
def test_mcmc_two_kernels(basic_lm: pm.Model):  # type: ignore
    interface = PyMCInterface(basic_lm, additional_vars=["sigma", "mu[0]"])
    state = interface.get_initial_state()
    builder = gs.EngineBuilder(1, 2)
    builder.set_initial_values(state)
    builder.set_model(interface)
    builder.set_duration(1000, 2000)

    builder.add_kernel(gs.NUTSKernel(["beta"]))
    builder.add_kernel(gs.NUTSKernel(["sigma_log__"]))

    builder.positions_included = ["sigma", "mu[0]"]

    engine = builder.build()

    engine.sample_all_epochs()
    results = engine.get_results()
    sum = gs.Summary(results)

    assert roughly_close(
        sum.quantities["mean"]["beta"], [0.90512399, 0.94801362, 2.0486103]
    )
    assert roughly_close(sum.quantities["mean"]["sigma_log__"], -0.01222878)
    assert roughly_close(sum.quantities["mean"]["sigma"], 0.99029397)
    assert roughly_close(sum.quantities["mean"]["mu[0]"], 0.53359132)



from collections.abc import Sequence
from dataclasses import dataclass
from typing import ClassVar

from liesel.goose.epoch import EpochState
from liesel.goose.kernel import (
    DefaultTransitionInfo,
    ModelMixin,
    TransitionOutcome,
    TuningOutcome,
    WarmupOutcome,
)
from liesel.goose.pytree import register_dataclass_as_pytree
from liesel.goose.types import Kernel, KeyArray, ModelInterface, ModelState, Position


@register_dataclass_as_pytree
@dataclass
class MockKernelState:
    pass

    @staticmethod
    def default() -> "MockKernelState":
        return MockKernelState()


@register_dataclass_as_pytree
@dataclass
class MockKernelTuningInfo:
    error_code: int
    time: int


@register_dataclass_as_pytree
@dataclass
class MockKernelTransInfo:
    error_code: int = 0
    acceptance_prob: float = 1
    position_moved: int = 1

    def minimize(self) -> DefaultTransitionInfo:
        return DefaultTransitionInfo(
            self.error_code, self.acceptance_prob, self.position_moved
        )


class MockKernel(ModelMixin):
    error_book: ClassVar[dict[int, str]] = {0: "no errors", 1: "error 1", 2: "error 2"}
    needs_history: ClassVar[bool] = False
    identifier: str = ""

    def __init__(
        self,
        position_keys: Sequence[str],
    ):
        self._model: None | ModelInterface = None
        self.position_keys = tuple(position_keys)

    def init_state(
        self, prng_key: KeyArray, model_state: ModelState
    ) -> MockKernelState:
        return MockKernelState.default()

    def start_epoch(
        self,
        prng_key: KeyArray,
        kernel_state: MockKernelState,
        model_state: ModelState,
        epoch: EpochState,
    ) -> MockKernelState:
        return kernel_state

    def end_epoch(
        self,
        prng_key: KeyArray,
        kernel_state: MockKernelState,
        model_state: ModelState,
        epoch: EpochState,
    ) -> MockKernelState:
        return kernel_state

    def transition(
        self,
        prng_key: KeyArray,
        kernel_state: MockKernelState,
        model_state: ModelState,
        epoch: EpochState,
    ) -> TransitionOutcome[MockKernelState, MockKernelTransInfo]:
        position = self.model.extract_position(self.position_keys, model_state)
        for key in position:
            position[key] = position[key] + 1.0

        new_model_state = self.model.update_state(position, model_state)

        info = MockKernelTransInfo()
        return TransitionOutcome(info, kernel_state, new_model_state)

    def tune(
        self,
        prng_key: KeyArray,
        kernel_state: MockKernelState,
        model_state: ModelState,
        epoch: EpochState,
        history: Position | None,
    ) -> TuningOutcome[MockKernelState, MockKernelTuningInfo]:
        info = MockKernelTuningInfo(error_code=0, time=epoch.time)
        return TuningOutcome(info, kernel_state)

    def end_warmup(
        self,
        prng_key: KeyArray,
        kernel_state: MockKernelState,
        model_state: ModelState,
        tuning_history: MockKernelTuningInfo | None,
    ) -> WarmupOutcome[MockKernelState]:
        return WarmupOutcome(0, kernel_state)


def typecheck() -> None:
    _: Kernel[MockKernelState, MockKernelTransInfo, MockKernelTuningInfo] = MockKernel(
        ["foo", "bar"]
    )



"""
Docstring utilities.
"""

from functools import wraps
from inspect import getmembers, isfunction


def _is_prop_or_method(x: object) -> bool:
    """
    Helper for selecting methods and properties of classes with
    :func:`inspect.getmembers`.

    Does not use :func:`inspect.ismethod`, because that returns *True* only for methods
    of class instances but not for the methods of the classes themselves.
    """
    return isfunction(x) or isinstance(x, property)


def usedocs(of: object, greedy: bool = True):
    """
    Fills missing docstrings of the decorated object with the docstrings from the
    specified object.

    Works on both functions and classes. Some notes on the details:

    1. By default, the decorator will paste the docstring from the object ``of`` to the
       decorated object.
    2. If the decorated object is a class, the decorator will also paste the docstrings
       of the methods and properties of the object ``of`` to their counterparts of the
       decorated object. This only works for methods and properties that have the same
       name for both objects. This behavior can be turned off by setting the argument
       ``greedy`` to ``False``.
    3. It is also possible to decorate individual methods and properties to reuse
       docstrings from specific objects.
    4. Existing docstrings will never be overwritten.

    Class and instance attribute docstrings are not covered by this decorator.

    Usage example::

        class ParentClass:
            '''Parent class documentation.'''

            def parent_method(self):
                '''Method documentation.'''

        @usedocs(ParentClass)
        class InheritingClass:

            def parent_method(self):
                pass
    """

    def _usedocs(obj):
        @wraps(obj)
        def wrapper():
            if not obj.__doc__:
                obj.__doc__ = of.__doc__

            if not greedy:
                return obj

            method_list = getmembers(obj, predicate=_is_prop_or_method)

            for name, method in method_list:
                if method.__doc__:
                    continue

                try:
                    parent_method = getattr(of, name)
                except AttributeError:
                    continue

                method.__doc__ = parent_method.__doc__

            return obj

        return wrapper()

    return _usedocs



"""
A :class:`.ModelInterface` for PyMC models.

To use this module, the PyMC package must be installed. To do so,
please install Liesel with the optional dependency PyMC:

.. code-block:: bash

    $ pip install liesel[pymc]

Example: A linear model
-----------------------

This model is also used in the tests::

    RANDOM_SEED = 123
    rng = np.random.RandomState(RANDOM_SEED)

    # set parameter values
    num_obs = 100
    sigma = 1.0
    beta = [1, 1, 2]

    # simulate covariates
    x1 = rng.randn(num_obs)
    x2 = 0.5 * rng.randn(num_obs)

    # simulate outcome variable
    y = beta[0] + beta[1] * x1 + beta[2] * x2 + sigma * rng.normal(size=num_obs)

    basic_model = pm.Model()
    with basic_model:
        # priors
        beta = pm.Normal("beta", mu=0, sigma=10, shape=3)
        sigma = pm.HalfNormal("sigma", sigma=1)
        # sigma is automatically transformed to real (log)
        # the new variable is called sigma_log__

        # predicted value
        mu = beta[0] + beta[1] * x1 + beta[2] * x2

        # track the predicted value of the first obs
        pm.Deterministic("mu[0]", mu[0])

        # distribution of response (likelihood)
        pm.Normal("y_obs", mu=mu, sigma=sigma, observed=y)

    interface = PyMCInterface(basic_model, additional_vars=["sigma", "mu[0]"])
    state = interface.get_initial_state()
    builder = gs.EngineBuilder(1, 2)
    builder.set_initial_values(state)
    builder.set_model(interface)
    builder.set_duration(1000, 2000)

    builder.add_kernel(gs.NUTSKernel(["beta"]))
    builder.add_kernel(gs.NUTSKernel(["sigma_log__"]))

    builder.positions_included = ["sigma", "mu[0]"]

    engine = builder.build()

    engine.sample_all_epochs()
    results = engine.get_results()
    sum = gs.Summary(results)
    sum

Transformations of RVs can be avoided by setting ``transform = None``
as a distribution argument.
"""

from __future__ import annotations

from collections.abc import Sequence

import jax

from liesel.goose.types import ModelState, Position

try:
    import pytensor

    if jax.config.read("jax_enable_x64"):
        pytensor.config.floatX = "float64"
    else:
        pytensor.config.floatX = "float32"

    import pymc as pm
    from pymc.sampling_jax import get_jaxified_graph, get_jaxified_logp
except ImportError as e:
    raise ImportError(
        f"PyMC must be installed to use this module. Original exception: {e}"
    )


class PyMCInterface:
    """
    An implementation of the Goose :class:`~liesel.goose.types.ModelInterface`
    to be used with a PyMC model.

    The initial position can be extracted with :meth:`.get_initial_state`.
    The model state is represented as a ``Position``.

    By default, only non-observed random variables are available via
    :meth:`.extract_position`. This includes transformed but not untransformed
    variables. Also, ``Deterministic``'s are not available. To make them trackable
    for the Goose :class:`~liesel.goose.engine.Engine`, these variables must be
    mentioned in the constructor.

    Parameters
    ----------
    model
        A PyMC model.
    additional_vars
        Variables that should be available via :meth:`.extract_position` \
        but are not by default.
    """

    def __init__(self, model: pm.Model, additional_vars: Sequence[str] = ()):
        self._pymc_model = model
        self._log_prob = get_jaxified_logp(self._pymc_model)
        self._rv_names = [rv.name for rv in model.value_vars]
        self._additional_vars = additional_vars

        # create a function to calculate the additional vars

        all_vars = pm.util.get_default_varnames(
            pm.modelcontext(model).unobserved_value_vars, include_transformed=True
        )
        selected_vars = [var for var in all_vars if var.name in self._additional_vars]
        self._calc_add_vars = get_jaxified_graph(
            inputs=model.value_vars, outputs=selected_vars
        )

    def get_initial_state(self) -> Position:
        """Returns the model's initial state."""
        return Position(self._pymc_model.initial_point())

    def extract_position(
        self, position_keys: Sequence[str], model_state: ModelState
    ) -> Position:
        """
        Extracts a sub-position specified by ``position_keys`` from ``model_state``.
        """
        # only extend the state if requested
        if self._additional_vars and any(
            key in position_keys for key in self._additional_vars
        ):
            model_state = model_state.copy()
            rv_values = [model_state[rv] for rv in self._rv_names]
            additional_values = self._calc_add_vars(*rv_values)

            for key, val in zip(self._additional_vars, additional_values):
                model_state[key] = val

        return Position({key: model_state[key] for key in position_keys})

    def update_state(self, position: Position, model_state: ModelState) -> ModelState:
        """Updates the model state with the position returning the new model state."""
        ms: Position = model_state.copy()  # do not change the input (escaped traces)
        ms.update(position)
        return ms

    def log_prob(self, model_state: ModelState) -> float:
        """Computes the unnormalized log-probability given the model state."""
        rv_values = [model_state[rv] for rv in self._rv_names]
        return self._log_prob(rv_values)



"""
Metroplis Hastings kernel. This kernel allows for a user-defined proposal functions and
adds the MH step. Optional, the kernel supports a stepsize adaptation.
"""

from collections.abc import Callable, Sequence
from typing import ClassVar, NamedTuple

import jax

from .da import da_finalize, da_init, da_step
from .epoch import EpochState
from .kernel import (
    DefaultTransitionInfo,
    DefaultTuningInfo,
    ModelMixin,
    TransitionMixin,
    TransitionOutcome,
    TuningOutcome,
    WarmupOutcome,
)
from .mh import mh_step
from .rw import RWKernelState
from .types import KeyArray, ModelState, Position, TuningInfo


class MHProposal(NamedTuple):
    position: Position
    log_correction: float
    """
    Let :math:`q(x' | x)` be the prosal density, then :math:`log(q(x'|x) / q(x | x'))`
    is the log_mh_correction.
    """


MHTransitionInfo = DefaultTransitionInfo
MHTuningInfo = DefaultTuningInfo
MHProposalFn = Callable[[KeyArray, ModelState, float], MHProposal]


class MHKernel(ModelMixin, TransitionMixin[RWKernelState, MHTransitionInfo]):
    """
    A Metropolis-Hastings kernel implementing the :class:`.Kernel` protocol.

    The user needs to provide a proposal function that proposes a new state and the
    log_correction.

    If ``da_tune_step_size`` is ``True`` the stepsize passed as an argument to the
    proposal function is tuned using the dual averging algorithm. Step size is tuned on
    the fly during all adaptive epochs.
    """

    error_book: ClassVar[dict[int, str]] = {0: "no errors", 90: "nan acceptance prob"}
    """Dict of error codes and their meaning."""
    needs_history: ClassVar[bool] = False
    """Whether this kernel needs its history for tuning."""
    identifier: str = ""
    """Kernel identifier, set by :class:`.EngineBuilder`"""
    position_keys: tuple[str, ...]
    """Tuple of position keys handled by this kernel."""

    def __init__(
        self,
        position_keys: Sequence[str],
        proposal_fn: MHProposalFn,
        initial_step_size: float = 1.0,
        da_tune_step_size=False,
        da_target_accept: float = 0.234,
        da_gamma: float = 0.05,
        da_kappa: float = 0.75,
        da_t0: int = 10,
    ):
        self._model = None
        self.position_keys = tuple(position_keys)
        self._proposal_fn = proposal_fn
        self.initial_step_size = initial_step_size
        self.da_tune_step_size = da_tune_step_size
        self.da_target_accept = da_target_accept
        self.da_gamma = da_gamma
        self.da_kappa = da_kappa
        self.da_t0 = da_t0

    def init_state(self, prng_key, model_state):
        """Initializes the kernel state."""

        return RWKernelState(step_size=self.initial_step_size)

    def _standard_transition(
        self,
        prng_key: KeyArray,
        kernel_state: RWKernelState,
        model_state: ModelState,
        epoch: EpochState,
    ) -> TransitionOutcome[RWKernelState, DefaultTransitionInfo]:
        """Performs an MCMC transition *without* dual averaging."""

        key, subkey = jax.random.split(prng_key)
        step_size = kernel_state.step_size

        # generate a proposal
        proposal = self._proposal_fn(key, model_state, step_size)

        # metropolis-hastings calibration
        info, model_state = mh_step(
            subkey,
            self.model,
            proposal.position,
            model_state,
            proposal.log_correction,
        )
        return TransitionOutcome(info, kernel_state, model_state)

    def _adaptive_transition(
        self,
        prng_key: KeyArray,
        kernel_state: RWKernelState,
        model_state: ModelState,
        epoch: EpochState,
    ) -> TransitionOutcome[RWKernelState, DefaultTransitionInfo]:
        """Performs an MCMC transition *with* dual averaging."""

        outcome = self._standard_transition(prng_key, kernel_state, model_state, epoch)

        if self.da_tune_step_size:
            da_step(
                outcome.kernel_state,
                outcome.info.acceptance_prob,
                epoch.time_in_epoch,
                self.da_target_accept,
                self.da_gamma,
                self.da_kappa,
                self.da_t0,
            )

        return outcome

    def tune(
        self,
        prng_key: KeyArray,
        kernel_state: RWKernelState,
        model_state: ModelState,
        epoch: EpochState,
        history: Position | None = None,
    ) -> TuningOutcome[RWKernelState, DefaultTuningInfo]:
        """Currently does nothing."""

        info = MHTuningInfo(error_code=0, time=epoch.time)
        return TuningOutcome(info, kernel_state)

    def start_epoch(
        self,
        prng_key: KeyArray,
        kernel_state: RWKernelState,
        model_state: ModelState,
        epoch: EpochState,
    ) -> RWKernelState:
        """Resets the state of the dual averaging algorithm."""

        da_init(kernel_state)
        return kernel_state

    def end_epoch(
        self,
        prng_key: KeyArray,
        kernel_state: RWKernelState,
        model_state: ModelState,
        epoch: EpochState,
    ) -> RWKernelState:
        """
        Sets the step size as found by the dual averaging algorithm.
        """

        da_finalize(kernel_state)
        return kernel_state

    def end_warmup(
        self,
        prng_key: KeyArray,
        kernel_state: RWKernelState,
        model_state: ModelState,
        tuning_history: TuningInfo | None,
    ) -> WarmupOutcome[RWKernelState]:
        """Currently does nothing."""

        return WarmupOutcome(error_code=0, kernel_state=kernel_state)



"""
Goose model interface.
"""

from __future__ import annotations

from collections.abc import Iterable

from liesel.goose.types import ModelState, Position

from .model import Model


class GooseModel:
    """
    A :class:`.ModelInterface` for a Liesel :class:`.Model`.

    Parameters
    ----------
    model
        A Liesel :class:`.Model`.
    """

    def __init__(self, model: Model):
        self._model = model._copy_computational_model()

    def extract_position(
        self, position_keys: Iterable[str], model_state: ModelState
    ) -> Position:
        """
        Extracts a position from a model state.

        Parameters
        ----------
        position_keys
            An iterable of variable or node names.
        model_state
            A dictionary of node names and their corresponding :class:`.NodeState`.
        """
        position = {}

        for key in position_keys:
            try:
                position[key] = model_state[key].value
            except KeyError:
                node_key = self._model.vars[key].value_node.name
                position[key] = model_state[node_key].value

        return Position(position)

    def update_state(self, position: Position, model_state: ModelState) -> ModelState:
        """
        Updates and returns a model state given a position.

        Parameters
        ----------
        position
            A dictionary of variable or node names and values.
        model_state
            A dictionary of node names and their corresponding :class:`.NodeState`.

        Warnings
        --------
        The ``model_state`` must be up-to-date, i.e. it must *not* contain any outdated
        nodes. Updates can only be triggered through new variable or node values in the
        ``position``. If you supply a ``model_state`` with outdated nodes, these nodes
        and their outputs will not be updated.
        """

        # sets all outdated flags in the model state to false
        # this is required to make the function jittable

        self._model.state = model_state

        for node in self._model.nodes.values():
            node._outdated = False

        for key, value in position.items():
            try:
                self._model.nodes[key].value = value  # type: ignore  # data node
            except KeyError:
                self._model.vars[key].value = value

        self._model.update()
        return self._model.state

    def log_prob(self, model_state: ModelState) -> float:
        """
        Returns the log-probability from a model state.

        Parameters
        ----------
        model_state
            A dictionary of node names and their corresponding :class:`.NodeState`.
        """
        return model_state["_model_log_prob"].value



"""
some tests for the engine
"""

from dataclasses import dataclass
from typing import ClassVar

import jax
import jax.numpy as jnp
import numpy as np

from liesel.goose.builder import EngineBuilder
from liesel.goose.chain import EpochChainManager
from liesel.goose.engine import (
    Engine,
    SamplingResults,
    _add_time_dimension,
    stack_for_multi,
)
from liesel.goose.epoch import EpochConfig, EpochState, EpochType
from liesel.goose.kernel import DefaultTransitionInfo
from liesel.goose.kernel_sequence import KernelSequence
from liesel.goose.models import DictModel
from liesel.goose.pytree import register_dataclass_as_pytree
from liesel.goose.types import Array, KeyArray, ModelInterface, ModelState
from liesel.option import Option

from .deterministic_kernels import DetCountingKernel, DetCountingKernelState


@register_dataclass_as_pytree
@dataclass
class FooQuant:
    error_code: int
    result: tuple[Array, Array]


class FooQauntGen:
    error_book: ClassVar[dict[int, str]] = {0: "no errors"}

    def __init__(self, identifier):
        self.identifier = identifier

    def set_model(self, model: ModelInterface):
        pass

    def has_model(self) -> bool:
        return False

    def generate(
        self, prng_key: KeyArray, model_state: ModelState, epoch: EpochState
    ) -> FooQuant:
        u = jax.random.normal(prng_key)
        return FooQuant(0, (u, model_state["x"]))


def test_add_time_dimension():
    def get_dims(t):
        return [t[0].shape, t[1][0].shape, t[1][1]["f"].shape]

    tree0 = jax.jit(lambda x: x)(
        (jnp.array([1.0, 2.0]), [jnp.zeros((3, 3)), {"f": jnp.array([1.0, 2.0])}])
    )
    tree3 = _add_time_dimension(tree0)
    dims3 = [
        (2, 1),
        (3, 1, 3),
        (2, 1),
    ]
    assert dims3 == get_dims(tree3)


def test_error_log():
    errs: np.ndarray = np.array([0, 0, 1, 0, 0, 0, 1, 1]).reshape((2, -1))
    ti = DefaultTransitionInfo(errs, np.zeros((2, 4)), np.zeros((2, 4), np.int8))
    tis = {"kern0": ti}

    em = EpochChainManager()
    em.advance_epoch(EpochConfig(EpochType.POSTERIOR, 4, 1, None))
    em.append(tis)
    em.combine_all()

    sr = SamplingResults(
        EpochChainManager(),
        em,
        Option.none(),
        Option.none(),
        Option.none(),
        Option.none(),
        Option.none(),
        Option.none(),
    )

    error_log = sr.get_error_log().unwrap()
    kel = error_log["kern0"]
    assert kel.kernel_ident == "kern0"
    assert np.array_equal(kel.transition, np.array([2, 3]))
    assert np.array_equal(kel.error_codes, np.array([[1, 0], [1, 1]]))


def t_test_engine():
    num_chains = 4
    epoch_configs = [
        EpochConfig(EpochType.INITIAL_VALUES, 1, 1, None),
        EpochConfig(EpochType.FAST_ADAPTATION, 50, 1, None),
        EpochConfig(EpochType.BURNIN, 50, 1, None),
        EpochConfig(EpochType.POSTERIOR, 100, 1, None),
    ]

    ms = {"x": jnp.array(1), "y": jnp.array(-1)}
    mss = stack_for_multi([ms for _ in range(num_chains)])
    con = DictModel(lambda ms: -0.5 * ms["x"] ** 2 - 0.5 * ms["y"])
    ker0 = DetCountingKernel(["x"], DetCountingKernelState.default())
    ker1 = DetCountingKernel(["y"], DetCountingKernelState.default())
    ker0.set_model(con)
    ker1.set_model(con)

    ks = KernelSequence([ker0, ker1])

    seeds = jax.random.split(jax.random.PRNGKey(0), num_chains)

    engine = Engine(
        seeds,
        mss,
        ks,
        epoch_configs,
        25,
        con,
        ["x"],
        minimize_transition_infos=False,
        store_kernel_states=True,
        quantity_generators=[FooQauntGen("foo"), FooQauntGen("bar")],
    )

    engine.sample_all_epochs()

    results: SamplingResults = engine.get_results()

    print(results.positions.combine_all())
    print(results.transition_infos.combine_all())
    print(results.kernel_states.combine_all())

    print(results.get_posterior_samples())
    print(results.get_tuning_times())

    print(results.generated_quantities.unwrap().combine_all().unwrap())


def t_test_engine_builder() -> None:
    builder = EngineBuilder(seed=1, num_chains=4)

    builder.set_epochs(
        [
            EpochConfig(EpochType.INITIAL_VALUES, 1, 1, None),
            EpochConfig(EpochType.FAST_ADAPTATION, 50, 1, None),
            EpochConfig(EpochType.BURNIN, 55, 10, None),
            EpochConfig(EpochType.POSTERIOR, 100, 10, None),
        ]
    )
    ms = {"x": jnp.array(1), "y": jnp.array(-1)}
    builder.set_initial_values(ms, multiple_chains=False)
    con = DictModel(lambda ms: -0.5 * ms["x"] ** 2 - 0.5 * ms["y"])
    builder.set_model(con)
    builder.add_kernel(DetCountingKernel(["x"], DetCountingKernelState.default()))
    builder.add_kernel(DetCountingKernel(["y"], DetCountingKernelState.default()))
    builder.add_quantity_generator(FooQauntGen("foo"))
    builder.add_quantity_generator(FooQauntGen("bar"))
    builder.positions_excluded = ["y"]
    engine = builder.build()

    engine.sample_all_epochs()
    results: SamplingResults = engine.get_results()

    # print(results.get_posterior_samples())
    # print(results.get_tuning_times())

    # print(results.generated_quantities.unwrap().combine_all().unwrap())
    # print(results.transition_infos.combine_all().unwrap())

    # test thinning worked
    assert results.get_posterior_samples()["x"].shape == (4, 10)
    assert results.get_samples()["x"].shape == (4, 66)
    assert results.generated_quantities.unwrap().combine_all().unwrap()["foo"].result[
        0
    ].shape == (4, 66)

    # test thinning is not applied to TIs
    assert results.transition_infos.combine_all().unwrap()[
        "kernel_01"
    ].error_code.shape == (4, 205)


if __name__ == "__main__":
    t_test_engine_builder()



import jax.numpy as jnp
import numpy as np

import liesel
from liesel.experimental.arviz import to_arviz_inference_data
from liesel.goose.engine import SamplingResults


def test_structure(result: SamplingResults):
    infdat = to_arviz_inference_data(result)
    assert infdat.groups() == ["posterior"]

    assert infdat["posterior"]["foo"].shape == (3, 250, 3)
    assert infdat["posterior"]["bar"].shape == (3, 250, 3, 5, 7)
    assert infdat["posterior"]["baz"].shape == (
        3,
        250,
    )


def test_structure_warmup(result: SamplingResults):
    infdat = to_arviz_inference_data(result, include_warmup=True)
    print(infdat)

    assert "posterior" in infdat.groups() and "warmup_posterior" in infdat.groups()

    assert infdat["warmup_posterior"]["foo"].shape == (3, 50, 3)
    assert infdat["warmup_posterior"]["bar"].shape == (3, 50, 3, 5, 7)
    assert infdat["warmup_posterior"]["baz"].shape == (
        3,
        50,
    )


def test_attributes(result: SamplingResults):
    infdat = to_arviz_inference_data(result)
    infdat["posterior"].attrs["inference_library"] == "liesel"
    infdat["posterior"].attrs["inference_library_version"] == liesel.__version__


def test_posterior_mean(result: SamplingResults):
    infdat = to_arviz_inference_data(result)
    means = infdat["posterior"].mean(["chain", "draw"])

    assert np.allclose(means["foo"], [175.5, 176.5, 177.5])

    assert np.allclose(means["bar"], 175.5 * jnp.ones((3, 5, 7)))

    assert means["baz"] == 176.5



from jax.random import PRNGKey

from liesel.goose.epoch import EpochConfig, EpochType
from liesel.goose.kernel_sequence import KernelSequence, KernelStates
from liesel.goose.models import DictModel
from liesel.goose.types import Kernel

from .deterministic_kernels import DetCountingKernel, DetCountingKernelState


def test_kernel_sequence() -> None:
    mstate = {"x": 0.0, "y": 0.0}
    kstate0 = DetCountingKernelState.default()
    kstate1 = DetCountingKernelState.default()
    model = DictModel(lambda state: 0.0)
    ker0: Kernel = DetCountingKernel(["x"], kstate0)
    ker1: Kernel = DetCountingKernel(["y"], kstate1)
    ker0.set_model(model)
    ker1.set_model(model)
    ker0.identifier = "foo"
    ker1.identifier = "bar"
    kseq = KernelSequence([ker0, ker1])

    configs = [
        EpochConfig(EpochType.FAST_ADAPTATION, 3, 1, None),
        EpochConfig(EpochType.POSTERIOR, 5, 1, None),
    ]

    prng_key = PRNGKey(0)
    kstates: KernelStates = kseq.init_states(prng_key, mstate)

    # epoch 0
    epoch = configs[0].to_state(0, 0)
    kstates = kseq.start_epoch(prng_key, kstates, mstate, epoch)
    for _ in range(epoch.config.duration):
        epoch.time_in_epoch += 1
        res = kseq.transition(prng_key, kstates, mstate, epoch)
        mstate = res.model_state
        kstates = res.kernel_states
    kstates = kseq.end_epoch(prng_key, kstates, mstate, epoch)
    tuning_res = kseq.tune(prng_key, kstates, mstate, epoch, None)
    kstates = tuning_res.kernel_states
    tune_infos = tuning_res.infos

    # finalize warmup
    assert not kstates[0].warmup_finialized
    assert not kstates[1].warmup_finialized
    fin_warm = kseq.end_warmup(prng_key, kstates, mstate, tune_infos)
    kstates = fin_warm.kernel_states
    assert kstates[0].warmup_finialized
    assert kstates[1].warmup_finialized

    # epoch 1
    epoch = configs[1].to_state(1, 3)
    kstates = kseq.start_epoch(prng_key, kstates, mstate, epoch)
    for _ in range(epoch.config.duration):
        epoch.time_in_epoch += 1
        res = kseq.transition(prng_key, kstates, mstate, epoch)
        mstate = res.model_state
        kstates = res.kernel_states

    assert kstates[0].in_epoch
    assert kstates[1].epoch_counter == 2
    kstates = kseq.end_epoch(prng_key, kstates, mstate, epoch)
    assert not kstates[0].in_epoch

    assert kstates[0].tune_counter == 1
    assert kstates[1].tune_counter == 1

    assert kstates[0].total_transitions == 8
    assert kstates[1].total_transitions == 8

    assert int(mstate["x"]) == 20_005
    assert int(mstate["y"]) == 20_005



"""
Distributional regression.
"""

from __future__ import annotations

import jax.numpy as jnp
import jax.random
import numpy as np
import tensorflow_probability.substrates.jax.distributions as tfd

from liesel.distributions import MultivariateNormalDegenerate
from liesel.goose import EngineBuilder, GibbsKernel, IWLSKernel
from liesel.option import Option

from .goose import GooseModel
from .legacy import (
    DesignMatrix,
    Hyperparameter,
    InverseLink,
    Predictor,
    RegressionCoef,
    Response,
    Smooth,
    SmoothingParam,
)
from .model import GraphBuilder, Model
from .nodes import Array, Bijector, Dist, Distribution, Group, NodeState, Var

matrix_rank = np.linalg.matrix_rank


class DistRegBuilder(GraphBuilder):
    """A model builder for distributional regression models."""

    _howto = "Smooths need to be added first, then the predictors, then the response"

    def __init__(self) -> None:
        super().__init__()

        self._smooths: dict[str, list[Var]] = {}
        self._distributional_parameters: dict[str, Var] = {}
        self._response: Option[Var] = Option(None)

    @property
    def response(self) -> Var:
        """The response node."""
        return self._response.expect(f"No response in {repr(self)}")

    def _smooth_name(self, name: str | None, predictor: str, prefix: str) -> str:
        """Generates a name for a smooth if the ``name`` argument is ``None``."""

        other_smooths = self._smooths[predictor]
        other_names = [node.name for node in other_smooths if node.name]
        prefix = predictor + "_" + prefix
        counter = 0

        while prefix + str(counter) in other_names:
            counter += 1

        if not name:
            name = prefix + str(counter)

        if name in other_names:
            raise RuntimeError(
                f"Smooth {repr(name)} already exists in {repr(self)} "
                f"for predictor {repr(predictor)}"
            )

        return name

    def add_p_smooth(
        self,
        X: Array,
        m: float,
        s: float,
        predictor: str,
        name: str | None = None,
    ) -> Group:
        """
        Adds a parametric smooth to the model builder.

        Parameters
        ----------
        X
            The design matrix.
        m
            The mean of the Gaussian prior.
        s
            The standard deviation of the Gaussian prior.
        predictor
            The name of the predictor to add the smooth to.
        name
            The name of the smooth.
        """

        try:
            self._smooths[predictor]
        except KeyError:
            self._smooths[predictor] = []

        name = self._smooth_name(name, predictor, "p")

        X_var = DesignMatrix(X, name=name + "_X")
        m_var = Hyperparameter(m, name=name + "_m")
        s_var = Hyperparameter(s, name=name + "_s")

        beta = np.zeros(np.shape(X)[-1], np.float32)
        beta_distribution = Dist(tfd.Normal, loc=m_var, scale=s_var)
        beta_var = RegressionCoef(beta, beta_distribution, name + "_beta")

        smooth_var = Smooth(X_var, beta_var, name=name)
        self._smooths[predictor].append(smooth_var)

        group = Group(name, smooth=smooth_var, beta=beta_var, X=X_var, m=m_var, s=s_var)

        self.add_groups(group)
        return group

    def add_np_smooth(
        self,
        X: Array,
        K: Array,
        a: float,
        b: float,
        predictor: str,
        name: str | None = None,
    ) -> Group:
        """
        Adds a non-parametric smooth to the model builder.

        Parameters
        ----------
        X
            The design matrix.
        K
            The penalty matrix.
        a
            The a, :math:`\\alpha` or concentration parameter of the inverse gamma
            prior.
        b
            The b, :math:`\\beta` or scale parameter of the inverse gamma prior.
        predictor
            The name of the predictor to add the smooth to.
        name
            The name of the smooth.
        """

        try:
            self._smooths[predictor]
        except KeyError:
            self._smooths[predictor] = []

        name = self._smooth_name(name, predictor, "np")

        X_var = DesignMatrix(X, name=name + "_X")
        K_var = Hyperparameter(K, name=name + "_K")
        a_var = Hyperparameter(a, name=name + "_a")
        b_var = Hyperparameter(b, name=name + "_b")

        rank_var = Hyperparameter(matrix_rank(K), name=name + "_rank")
        tau2_distribution = Dist(tfd.InverseGamma, concentration=a_var, scale=b_var)
        tau2_var = SmoothingParam(10000.0, tau2_distribution, name + "_tau2")

        beta = np.zeros(np.shape(X)[-1], np.float32)
        beta_distribution = Dist(
            MultivariateNormalDegenerate.from_penalty,
            loc=0.0,
            var=tau2_var,
            pen=K_var,
            rank=rank_var,
        )
        beta_var = RegressionCoef(beta, beta_distribution, name + "_beta")

        smooth_var = Smooth(X_var, beta_var, name=name)
        self._smooths[predictor].append(smooth_var)

        group = Group(
            name,
            smooth=smooth_var,
            beta=beta_var,
            tau2=tau2_var,
            rank=rank_var,
            X=X_var,
            K=K_var,
            a=a_var,
            b=b_var,
        )

        self.add_groups(group)
        return group

    def add_predictor(self, name: str, inverse_link: type[Bijector]) -> DistRegBuilder:
        """
        Adds a predictor to the model builder.

        Parameters
        ----------
        name
            The name of the parameter of the response distribution. Must match the name
            of the parameter of the TFP distribution.
        inverse_link
            The inverse link mapping the regression predictor to the parameter of the
            response distribution. Either a string identifying a TFP bijector, or
            alternatively, a TFP-compatible bijector class. If a class is provided
            instead of a string, the user needs to make sure it uses the right NumPy
            implementation.
        """

        if name not in self._smooths or not self._smooths[name]:
            msg = f"No smooths in {repr(self)} for predictor {repr(name)}. "
            raise RuntimeError(msg + self._howto)

        smooth_vars = self._smooths[name]
        predictor_var = Predictor(name=name + "_pdt", *smooth_vars)
        parameter_var = InverseLink(predictor_var, inverse_link, name=name)
        self._distributional_parameters[name] = parameter_var
        self.add(predictor_var, parameter_var)

        return self

    def add_response(
        self, response: Array, distribution: type[Distribution]
    ) -> DistRegBuilder:
        """
        Adds the response to the model builder.

        Parameters
        ----------
        response
            The response vector or matrix.
        distribution
            The conditional distribution of the response variable. Either a string
            identifying a TFP distribution, or alternatively, a TFP-compatible
            distribution class. If a class is provided instead of a string, the user
            needs to make sure it uses the right NumPy implementation.
        """

        if not self._distributional_parameters:
            raise RuntimeError(f"No predictors in {repr(self)}. {self._howto}")

        response_distribution = Dist(
            distribution, **self._distributional_parameters  # type: ignore
        )

        response_var = Response(response, response_distribution, "response")
        self._response = Option(response_var)
        self.add(response_var)

        return self


def tau2_gibbs_kernel(group: Group) -> GibbsKernel:
    """Builds a Gibbs kernel for a smoothing parameter with an inverse gamma prior."""
    position_key = group["tau2"].name

    def transition(prng_key, model_state: dict[str, NodeState]):
        a_prior = group.value_from(model_state, "a")
        rank = group.value_from(model_state, "rank")

        a_gibbs = jnp.squeeze(a_prior + 0.5 * rank)

        b_prior = group.value_from(model_state, "b")
        beta = group.value_from(model_state, "beta")
        K = group.value_from(model_state, "K")

        b_gibbs = jnp.squeeze(b_prior + 0.5 * (beta @ K @ beta))

        draw = b_gibbs / jax.random.gamma(prng_key, a_gibbs)

        return {position_key: draw}

    return GibbsKernel([position_key], transition)


def dist_reg_mcmc(model: Model, seed: int, num_chains: int) -> EngineBuilder:
    """
    Configures an :class:`.EngineBuilder` for a distributional regression model.

    The EngineBuilder uses a Metropolis-in-Gibbs MCMC algorithm with an
    :class:`.IWLSKernel` for the regression coefficients and a :class:`.GibbsKernel` for
    the smoothing parameters for a distributional regression model.

    Parameters
    ----------
    model
        A model built with a :class:`.DistRegBuilder`.
    seed
        The PRNG seed for the engine builder.
    num_chains
        The number of chains to be sampled.
    """

    builder = EngineBuilder(seed, num_chains)

    builder.set_model(GooseModel(model))
    builder.set_initial_values(model.state)

    for group in model.groups().values():
        if "tau2" in group:
            tau2_kernel = tau2_gibbs_kernel(group)  # type: ignore  # only vars
            builder.add_kernel(tau2_kernel)

        if "beta" in group:
            position_key = group["beta"].name
            beta_kernel = IWLSKernel([position_key])
            builder.add_kernel(beta_kernel)

    return builder



import jax.numpy as jnp
import numpy as np

from liesel.goose.engine import SamplingResults
from liesel.goose.summary_m import Summary

# TODO: add tests to test correctness of quantities
# TODO: speed up tests


def test_shapes(result: SamplingResults):
    summary = Summary(result)

    # combined chains
    assert summary.quantities["mean"]["foo"].shape == (3,)
    assert summary.quantities["mean"]["bar"].shape == (3, 5, 7)
    assert summary.quantities["mean"]["baz"].shape == ()

    assert summary.quantities["quantile"]["foo"].shape == (
        3,
        3,
    )
    assert summary.quantities["quantile"]["bar"].shape == (3, 3, 5, 7)
    assert summary.quantities["quantile"]["baz"].shape == (3,)

    assert summary.quantities["hdi"]["foo"].shape == (
        2,
        3,
    )
    assert summary.quantities["hdi"]["bar"].shape == (2, 3, 5, 7)
    assert summary.quantities["hdi"]["baz"].shape == (2,)

    # combined chains
    summary = Summary(result, quantiles=(0.2, 0.4), per_chain=True)

    assert summary.quantities["mean"]["foo"].shape == (
        3,
        3,
    )
    assert summary.quantities["mean"]["bar"].shape == (3, 3, 5, 7)
    assert summary.quantities["mean"]["baz"].shape == (3,)

    assert summary.quantities["quantile"]["foo"].shape == (
        3,
        2,
        3,
    )
    assert summary.quantities["quantile"]["bar"].shape == (3, 2, 3, 5, 7)
    assert summary.quantities["quantile"]["baz"].shape == (3, 2)

    assert summary.quantities["hdi"]["foo"].shape == (
        3,
        2,
        3,
    )
    assert summary.quantities["hdi"]["bar"].shape == (3, 2, 3, 5, 7)
    assert summary.quantities["hdi"]["baz"].shape == (
        3,
        2,
    )


def test_additional_chain(result: SamplingResults):
    chain = result.get_posterior_samples()
    chain["expbaz"] = jnp.log(chain["baz"] + 1)
    chain["expbar"] = jnp.log(chain["bar"] + 1)

    summary = Summary(result, chain)

    assert summary.quantities["mean"]["expbar"].shape == (3, 5, 7)
    assert summary.quantities["mean"]["expbaz"].shape == ()


def test_selected(result: SamplingResults):
    summary = Summary(result, selected=["foo"])

    assert "foo" in summary.quantities["mean"]
    assert "bar" not in summary.quantities["mean"]
    assert "baz" not in summary.quantities["mean"]


def test_deselected(result: SamplingResults):
    summary = Summary(result, deselected=["baz"])

    assert "foo" in summary.quantities["mean"]
    assert "bar" in summary.quantities["mean"]
    assert "baz" not in summary.quantities["mean"]


def test_mean(result: SamplingResults):
    summary = Summary(result)
    assert jnp.allclose(
        summary.quantities["mean"]["foo"], jnp.array([175.5, 176.5, 177.5])
    )

    assert jnp.allclose(summary.quantities["mean"]["bar"], 175.5 * jnp.ones((3, 5, 7)))

    assert jnp.allclose(summary.quantities["mean"]["baz"], jnp.array(176.5))


def test_config(result: SamplingResults):
    summary = Summary(result, quantiles=(0.4, 0.6), hdi_prob=0.5)
    assert summary.config["chains_merged"]
    assert summary.config["quantiles"] == (0.4, 0.6)
    assert summary.config["hdi_prob"] == 0.5


def test_sample_info(result: SamplingResults):
    summary = Summary(result)
    print(summary.sample_info)
    assert summary.sample_info["num_chains"] == 3
    assert summary.sample_info["sample_size_per_chain"] == 250


def test_df_sample_info(result: SamplingResults):
    summary = Summary(result, selected=["baz"]).to_dataframe()
    assert summary["sample_size"][0] == 3 * 250

    summary = Summary(result, per_chain=True, selected=["baz"]).to_dataframe()
    assert summary["sample_size"][0] == 250


def test_error_summary(result: SamplingResults):
    # add some error codes to the chain

    # epoch 1 - warmup
    ecs = np.array(
        result.transition_infos.get_specific_chain(1)
        ._chunks_list[0]["kernel_00"]
        .error_code
    )
    ecs[0, 0] = 1
    ecs[:, 3] = 1
    ecs[2, 3:5] = 2
    result.transition_infos.get_specific_chain(1)._chunks_list[0][
        "kernel_00"
    ].error_code = ecs

    # posterior
    ecs = np.array(
        result.transition_infos.get_current_chain()
        ._chunks_list[0]["kernel_00"]
        .error_code
    )
    ecs[0, 0] = 1
    ecs[:, 230] = 1
    ecs[1, 220:225] = 2
    result.transition_infos.get_current_chain()._chunks_list[0][
        "kernel_00"
    ].error_code = ecs

    # create the summary object
    summary = Summary(result, selected=["baz"])
    es = summary.error_summary["kernel_00"]

    # check error_code correspond
    assert es[1].error_code == 1
    assert es[2].error_code == 2

    # check that these error codes don't have entries
    assert 3 not in es
    assert 0 not in es

    # check that error messages are correct
    assert es[1].error_msg == "error 1"
    assert es[2].error_msg == "error 2"

    # check that counts are correct - overall
    assert np.all(es[1].count_per_chain == np.array([4, 2, 1]))
    assert np.all(es[2].count_per_chain == np.array([0, 5, 2]))

    # check that counts are correct - only in the posterior
    assert np.all(es[1].count_per_chain_posterior == np.array([2, 1, 1]))
    assert np.all(es[2].count_per_chain_posterior == np.array([0, 5, 0]))


def test_single_chain_repr_fs_return(single_chain_result: SamplingResults):
    summary = Summary(single_chain_result)
    md = summary._repr_markdown_()
    html = summary._repr_html_()
    assert isinstance(md, str)
    assert isinstance(html, str)


def test_per_chain_quantiles(result: SamplingResults):
    summary = Summary(result, per_chain=True)
    cols = [
        "var_fqn",
        "chain_index",
        "mean",
        "q_0.05",
        "q_0.5",
        "q_0.95",
        "hdi_low",
        "hdi_high",
    ]
    df = summary.to_dataframe().loc["baz"][cols]

    assert np.allclose(df["q_0.05"], 64.449997)
    assert np.allclose(df["q_0.5"], 176.5)
    assert np.allclose(df["q_0.95"], 288.549988)
    assert np.allclose(df["hdi_low"], 52.0)
    assert np.allclose(df["hdi_high"], 277.0)


def test_quantity_shape(result_for_quants: SamplingResults):
    """
    Confirms that the HDI and quantile quantities at the summary object have the
    intended shape:

    - First index refers to the chain
    - Second index refers to the quantile/hdi
    - Following indices refer to individual parameters
    """
    summary = Summary(result_for_quants, per_chain=True)

    assert summary.quantities["quantile"]["foo"].shape == (4, 3, 5)
    assert summary.quantities["quantile"]["bar"].shape == (4, 3, 3, 5, 7)
    assert summary.quantities["quantile"]["baz"].shape == (4, 3)

    assert summary.quantities["hdi"]["foo"].shape == (4, 2, 5)
    assert summary.quantities["hdi"]["bar"].shape == (4, 2, 3, 5, 7)
    assert summary.quantities["hdi"]["baz"].shape == (4, 2)



"""
Standard warmup schemes.
"""

from functools import partial

from .epoch import EpochConfig, EpochType

_EpochConfig = partial(EpochConfig, optional=None)


def stan_epochs(
    warmup_duration: int = 1000,
    posterior_duration: int = 1000,
    init_duration: int = 75,
    term_duration: int = 50,
    base_duration: int = 25,
    thinning_posterior: int = 1,
    thinning_warmup: int = 1,
) -> list[EpochConfig]:
    """
    Sets up a list of :class:`.EpochConfig`'s.

    This follows the Stan Development Team (Stan Reference Manual 2021, Chapter 15.2
    [#stanmanual]_).

    Parameters
    ----------
    warmup_duration
        The number of warmup samples.
    posterior_duration
        The number of posterior samples.
    init_duration
        The number of samples in the *initial fast* adaptation epoch.
    term_duration
        The number of samples in the *final fast* adaptation epoch.
    base_duration
        The number of samples in the *first slow* adaptation epoch.
    thinning_posterior
        Thinning applied in the posterior epoch.
    thinning_warmup
        Thinning applied in each warmup.

    Warnings
    --------
    Kernels which rely on history tuning might not be able to deal with thinning during
    warmup.

    References
    ----------

    .. [#stanmanual] https://mc-stan.org/docs/2_28/reference-manual/hmc-algorithm-parameters.html # noqa: E501

    """

    if warmup_duration < 20:
        raise ValueError("warmup_duration too short (< 20)")

    if warmup_duration < init_duration + term_duration + base_duration:
        raise ValueError(
            "warmup_duration too short "
            "(< init_duration + term_duration + base_duration)"
        )

    epochs = [_EpochConfig(EpochType.INITIAL_VALUES, duration=1, thinning=1)]
    epochs.append(
        _EpochConfig(EpochType.FAST_ADAPTATION, init_duration, thinning_warmup)
    )

    time_left = warmup_duration - init_duration - term_duration
    this_time = base_duration

    while 3 * this_time <= time_left:
        epochs.append(
            _EpochConfig(EpochType.SLOW_ADAPTATION, this_time, thinning_warmup)
        )
        time_left -= this_time
        this_time *= 2

    epochs.append(_EpochConfig(EpochType.SLOW_ADAPTATION, time_left, thinning_warmup))
    epochs.append(
        _EpochConfig(EpochType.FAST_ADAPTATION, term_duration, thinning_warmup)
    )
    epochs.append(
        _EpochConfig(EpochType.POSTERIOR, posterior_duration, thinning_posterior)
    )

    return epochs



import jax.numpy as jnp
import pytest

from liesel.goose.models import DictModel
from liesel.goose.types import ModelInterface, Position


def log_prob(state) -> float:
    x = state["x"]
    return -jnp.sum(x**2)


def create_state():
    return {"x": jnp.array([-1.0, 0.0, 1.0]), "y": 1}


def test_log_prob() -> None:
    conn: ModelInterface = DictModel(log_prob)
    state = create_state()
    lp = float(conn.log_prob(state))
    assert lp == pytest.approx(-2.0)


def test_extract() -> None:
    conn = DictModel(log_prob)
    state = create_state()

    pos1 = conn.extract_position(["x"], state)
    assert pos1["x"] == pytest.approx(state["x"])
    assert len(pos1) == 1

    pos2 = conn.extract_position(["x", "y"], state)
    assert pos2["x"] == pytest.approx(state["x"])
    assert pos2["y"] == state["y"]
    assert len(pos2) == len(state)
    assert list(pos2.keys()) == ["x", "y"]

    pos3 = conn.extract_position(["y", "x"], state)
    assert list(pos3.keys()) == ["y", "x"]


def test_update() -> None:
    conn = DictModel(log_prob)
    state0 = create_state()

    state1 = conn.update_state(Position({"y": 2}), state0)
    assert state1["y"] == 2
    assert state0["y"] == 1

    state2 = conn.update_state(Position({"y": 3, "x": 2 * state0["x"]}), state0)
    assert state2["y"] == 3
    assert state0["y"] == 1

    assert conn.log_prob(state0) == pytest.approx(-2.0)
    assert conn.log_prob(state2) == pytest.approx(-8.0)



from jax.random import PRNGKey

from liesel.goose.epoch import EpochConfig, EpochType
from liesel.goose.models import DictModel
from liesel.goose.types import Kernel

from .deterministic_kernels import (
    DetCountingKernel,
    DetCountingKernelState,
    DetCountingKernelTuningInfo,
    DetCountingTransInfo,
)


def test_counting_kernel() -> None:
    mstate = {"x": 0}
    kstate = DetCountingKernelState.default()
    model = DictModel(lambda state: 0.0)
    kernel: Kernel[
        DetCountingKernelState, DetCountingTransInfo, DetCountingKernelTuningInfo
    ] = DetCountingKernel(["x"], kstate)
    kernel.set_model(model)

    configs = [
        EpochConfig(EpochType.FAST_ADAPTATION, 3, 1, None),
        EpochConfig(EpochType.FAST_ADAPTATION, 5, 1, None),
        EpochConfig(EpochType.POSTERIOR, 7, 1, None),
    ]

    prng_key = PRNGKey(0)
    kstate = kernel.init_state(prng_key, mstate)

    # epoch 0
    epoch = configs[0].to_state(0, 0)
    kstate = kernel.start_epoch(prng_key, kstate, mstate, epoch)
    for _ in range(epoch.config.duration):
        epoch.time_in_epoch += 1
        res = kernel.transition(prng_key, kstate, mstate, epoch)
        mstate = res.model_state
        kstate = res.kernel_state
    kstate = kernel.end_epoch(prng_key, kstate, mstate, epoch)
    tuning_res = kernel.tune(prng_key, kstate, mstate, epoch, None)
    kstate = tuning_res.kernel_state
    _ = tuning_res.info

    # epoch 1
    epoch = configs[1].to_state(1, 3)
    kstate = kernel.start_epoch(prng_key, kstate, mstate, epoch)
    for _ in range(epoch.config.duration):
        epoch.time_in_epoch += 1
        res = kernel.transition(prng_key, kstate, mstate, epoch)
        mstate = res.model_state
        kstate = res.kernel_state
    kstate = kernel.end_epoch(prng_key, kstate, mstate, epoch)
    tuning_res = kernel.tune(prng_key, kstate, mstate, epoch, None)
    kstate = tuning_res.kernel_state
    tune_info1 = tuning_res.info

    # finalize warmup
    assert not kstate.warmup_finialized
    fin_warm = kernel.end_warmup(prng_key, kstate, mstate, tune_info1)
    kstate = fin_warm.kernel_state
    assert kstate.warmup_finialized
    assert kstate.total_adaptive_transitions == 8

    # epoch 2
    epoch = configs[2].to_state(2, 8)
    kstate = kernel.start_epoch(prng_key, kstate, mstate, epoch)
    for _ in range(epoch.config.duration):
        epoch.time_in_epoch += 1
        res = kernel.transition(prng_key, kstate, mstate, epoch)
        mstate = res.model_state
        kstate = res.kernel_state

    assert kstate.in_epoch
    assert kstate.epoch_counter == 3
    kstate = kernel.end_epoch(prng_key, kstate, mstate, epoch)
    assert not kstate.in_epoch

    assert kstate.tune_counter == 2
    assert kstate.total_transitions == 15
    assert kstate.total_adaptive_transitions == 8
    assert int(mstate["x"]) == 30_007



import tensorflow_probability.substrates.numpy.distributions as nd
from tensorflow_probability.python.internal import reparameterization


class NoDistribution(nd.Distribution):
    def __init__(self):
        super().__init__(
            dtype=float,
            reparameterization_type=reparameterization.FULLY_REPARAMETERIZED,
            validate_args=False,
            allow_nan_stats=True,
        )

    def _default_event_space_bijector(self):
        return None



"""
Diagnostic plots of the posterior samples.
"""

from collections.abc import Sequence
from typing import Any

import arviz
import jax.numpy as jnp
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

from liesel.goose.engine import SamplingResults


def _raise_chain_indices_error(
    chain_indices: Sequence[int], num_original_chains: int
) -> None:
    """Display informative error message with valid ``chain_indices`` inputs."""
    if any(
        chain_index not in range(num_original_chains) for chain_index in chain_indices
    ):
        raise ValueError(
            f"All chain indices must be between 0 and {num_original_chains-1} "
            "(bounds inclusive)."
        )


def _validate_chain_indices(
    chain_indices: int | Sequence[int] | None,
    num_original_chains: int,
) -> Sequence[int]:
    """
    Convert ``int`` or ``None`` input of ``chain_indices`` to sequence of integers.
    """
    if chain_indices is None:
        return list(range(num_original_chains))

    if isinstance(chain_indices, int):
        chain_indices = [chain_indices]

    _raise_chain_indices_error(chain_indices, num_original_chains)
    return chain_indices


def _raise_dimension_error(param: str, num_dim: int) -> None:
    """Check for correct array dimensions of posterior samples."""
    if num_dim not in (2, 3):
        raise ValueError(
            f"Array of posterior samples for {param} has the wrong number of"
            f"dimensions.\nExpected 2 or 3, got {num_dim}."
        )


def _adjust_dimensions(param_chains: np.ndarray, num_dim: int) -> np.ndarray:
    """
    Make shape of posterior samples for one dimensional parameters (e.g. ``log_sigma``)
    consistent with multi-dimensional parameters.
    """
    if num_dim == 2:
        param_chains = np.expand_dims(param_chains, axis=-1)
    return param_chains


def _raise_param_indices_error(
    param_indices: Sequence[int], num_original_subparams: int, param: str
) -> None:
    """
    Display informative error message with valid ``param_indices`` inputs for this
    specific ``param``.
    """
    if any(
        param_index not in range(num_original_subparams)
        for param_index in param_indices
    ):
        raise ValueError(
            f"All param indices for {param} must be between "
            f"0 and {num_original_subparams-1} (bounds inclusive)."
        )


def _validate_param_indices(
    param_indices: int | Sequence[int] | None, num_original_subparams: int, param: str
) -> Sequence[int]:
    """
    Convert ``int`` or ``None`` input of ``param_indices`` to sequence of integers.
    """
    if param_indices is None:
        return list(range(num_original_subparams))

    if isinstance(param_indices, int):
        param_indices = [param_indices]

    _raise_param_indices_error(param_indices, num_original_subparams, param)

    return param_indices


def _move_col_first(df: pd.DataFrame, colname: str) -> pd.DataFrame:
    """Move last column of a :class:`~pandas.DataFrame` to the first column."""
    return df[[colname] + [col for col in df.columns if col != colname]]


def _validate_params(
    posterior_samples: dict[str, jnp.DeviceArray], params: str | list[str] | None
) -> list[str]:
    """Convert ``str`` or ``None`` input of ``params`` to sequence of strings."""
    posterior_keys = list(posterior_samples.keys())
    if params is None:
        return posterior_keys

    if isinstance(params, str):
        params = [params]

    if any(param not in posterior_keys for param in params):
        raise KeyError(f"All params must be in {posterior_keys}.")

    return params


def _subparam_chains_to_df(
    subparam_chains: np.ndarray, param_index: int
) -> pd.DataFrame:
    """
    Convert array of posterior samples for a single subparameter (e.g. ``beta[0]``) to a
    pandas data frame.
    """

    subparam_df = (
        pd.DataFrame(subparam_chains)
        .melt(ignore_index=False)
        .rename(columns={"variable": "iteration"})
        .reset_index()
        .rename(columns={"index": "chain_index"})
        .sort_values(by=["chain_index", "iteration"], ignore_index=True)
        .assign(param_index=param_index)
    )

    return _move_col_first(subparam_df, colname="param_index")


def _preprocess_param_chains(
    posterior_samples: dict[str, jnp.DeviceArray], param: str
) -> np.ndarray:
    """Convert array of posteror samples for each parameter to equal dimensions."""

    param_chains = np.array(posterior_samples[param])
    num_dim = param_chains.ndim

    _raise_dimension_error(param, num_dim)
    param_chains = _adjust_dimensions(param_chains, num_dim)
    return param_chains


def _convert_to_sequence(indices: int | Sequence[int]) -> Sequence[int]:
    """Convert integer parameter and chain indices to list or tuple."""

    if isinstance(indices, int):
        indices = [indices]

    return indices


def _filter_param_df(
    param_df: pd.DataFrame,
    param_indices: int | Sequence[int] | None,
    chain_indices: int | Sequence[int] | None,
    max_chains: int | None,
) -> pd.DataFrame:
    """
    Filters the plotting data to contain only the specified parameter and chain indices.
    Output data contains not more than `max_chains` chain indices.
    """

    if chain_indices is not None:
        chain_indices = _convert_to_sequence(chain_indices)
        param_df = param_df.loc[param_df["chain_index"].isin(chain_indices)]

    if param_indices is not None:
        param_indices = _convert_to_sequence(param_indices)
        param_df = param_df.loc[param_df["param_index"].isin(param_indices)]

    if max_chains is not None and param_df["chain_index"].nunique() > max_chains:
        last_chain_index = sorted(param_df["chain_index"].unique())[max_chains - 1]
        param_df = param_df.loc[param_df["chain_index"] <= last_chain_index]

    return param_df


def _postprocess_param_df(
    param_df: pd.DataFrame,
    param: str,
    param_indices: int | Sequence[int] | None,
    chain_indices: int | Sequence[int] | None,
    max_chains: int | None,
) -> pd.DataFrame:
    """
    Combines parameter and parameter index column and filters chain and parameter
    indices.
    """

    num_original_chains = param_df["chain_index"].nunique()
    num_original_subparams = param_df["param_index"].nunique()

    if num_original_subparams > 1:
        name = param_df["param"].astype(str)
        index = param_df["param_index"].astype(str)
        param_df.loc[:, "param_label"] = name + "[" + index + "]"
    else:
        name = param_df["param"].astype(str)
        param_df["param_label"] = name

    chain_indices = _validate_chain_indices(chain_indices, num_original_chains)
    param_indices = _validate_param_indices(
        param_indices, num_original_subparams, param
    )
    param_df = _filter_param_df(param_df, param_indices, chain_indices, max_chains)

    return param_df


def _collect_subparam_dfs(
    posterior_samples: dict[str, jnp.DeviceArray],
    param: str,
    param_indices: int | Sequence[int] | None,
    chain_indices: int | Sequence[int] | None,
    max_chains: int | None,
) -> pd.DataFrame:
    """
    Combines individual data frames for each subparameter into a single data frame for
    each parameter.
    """

    param_chains = _preprocess_param_chains(posterior_samples, param)

    param_df = (
        pd.concat(
            [
                _subparam_chains_to_df(param_chains[..., param_index], param_index)
                for param_index in range(param_chains.shape[-1])
            ]
        )
        .assign(param=param)
        .reset_index(drop=True)
    )

    param_df = _move_col_first(param_df, colname="param")

    return _postprocess_param_df(
        param_df, param, param_indices, chain_indices, max_chains
    )


def _collect_param_dfs(
    results: SamplingResults,
    params: str | list[str] | None = None,
    param_indices: int | Sequence[int] | None = None,
    chain_indices: int | Sequence[int] | None = None,
    max_chains: int | None = 5,
    include_warmup: bool = False,
) -> pd.DataFrame:
    """Combines individual data frames for each parameter into a single data frame."""

    if include_warmup:
        samples = results.get_samples()
    else:
        samples = results.get_posterior_samples()
    params = _validate_params(samples, params)

    return pd.concat(
        [
            _collect_subparam_dfs(
                samples, param, param_indices, chain_indices, max_chains
            )
            for param in params
        ]
    ).reset_index(drop=True)


def _setup_plot_df(
    results: SamplingResults,
    params: str | list[str] | None,
    param_indices: int | Sequence[int] | None,
    chain_indices: int | Sequence[int] | None,
    max_chains: int | None,
    include_warmup: bool = False,
) -> pd.DataFrame:
    """Provides data input for all plotting functions."""

    return _collect_param_dfs(
        results,
        params,
        param_indices,
        chain_indices,
        max_chains,
        include_warmup,
    ).astype({"chain_index": "category"})


def _setup_scatterplot_df(
    results: SamplingResults,
    params: str | list[str] | None,
    param_indices: int | Sequence[int] | None,
    chain_indices: int | Sequence[int] | None,
    max_chains: int | None,
    include_warmup: bool = False,
) -> pd.DataFrame:
    """
    Provides bespoke data input for plot_scatter. If two indices *and* two params are
    specified, the first index refers the first param and the second index to the
    second.
    """
    if (
        isinstance(params, str)
        or params is None
        or isinstance(param_indices, int)
        or param_indices is None
    ):
        return _setup_plot_df(
            results, params, param_indices, chain_indices, max_chains, include_warmup
        )
    if not isinstance(params, str) and len(params) > 1:
        param0_df = _setup_plot_df(
            results,
            params[0],
            param_indices[0],
            chain_indices,
            max_chains,
            include_warmup,
        )
        param1_df = _setup_plot_df(
            results,
            params[1],
            param_indices[1],
            chain_indices,
            max_chains,
            include_warmup,
        )
        plot_df = pd.concat([param0_df, param1_df], ignore_index=True)
    else:
        plot_df = _setup_plot_df(
            results, params, param_indices, chain_indices, max_chains, include_warmup
        )
    return plot_df


def _set_plot_cols(plot_df: pd.DataFrame, ncol: int) -> int:
    """Determines number of facets within each row of the grid."""

    num_subparams = plot_df["param_label"].nunique()
    return min(ncol, num_subparams)


def _set_aesthetics(
    g: sns.FacetGrid, title: str | None, title_spacing: float, xlabel: str, ylabel: str
) -> sns.FacetGrid:
    """Adds titles, labels and correct spacing between facets."""

    g.set_titles(col_template="{col_name}")
    g.set_axis_labels(x_var=xlabel, y_var=ylabel)
    g.tight_layout()

    g.legend.set_title("Chain")

    if title is not None:
        g.fig.suptitle(title)
        g.fig.subplots_adjust(top=title_spacing)

    return g


def save_figure(g: sns.FacetGrid | None = None, save_path: str | None = None) -> None:
    """Saves plot to file."""

    if save_path is not None:
        if g is not None:
            g.fig.savefig(save_path)
        else:
            plt.savefig(save_path)
    else:
        plt.show()


def plot_trace(
    results: SamplingResults,
    params: str | list[str] | None = None,
    param_indices: int | Sequence[int] | None = None,
    chain_indices: int | Sequence[int] | None = None,
    max_chains: int | None = 5,
    title: str | None = None,
    title_spacing: float = 0.85,
    xlabel: str = "Iteration",
    style: str = "whitegrid",
    color_palette: str | list[str] | dict[int, str] | None = None,
    ncol: int = 3,
    height: int = 3,
    aspect_ratio: int = 1,
    save_path: str | None = None,
    include_warmup: bool = False,
    **kwargs,
) -> sns.FacetGrid:
    """
    Visualizes posterior samples over time with a trace plot.

    Parameters
    ----------
    results
        Result object of the sampling process. Must have a method
        ``get_posterior_samples()`` which extracts all samples from the posterior
        distribution.
    params
        Names of the model parameters that are contained in the plot. Must coincide
        with the dictionary keys of the `Position` with the posterior samples. If
        `None`, all parameters are included.
    param_indices
        Indices of each model parameter that are contained in the plot. Selects e.g.
        ``beta[0]`` out of a ``beta`` parameter vector. A single index can be specified
        as an integer or a sequence containing one integer. If ``None``, all
        subparameters are included.
    chain_indices
        Indices of chains for each model subparameter that are contained in the plot.
        Selects e.g. chain 0 and chain 2 out of multiple chains. A single index can be
        specified as an integer or a sequence containing one integer. If ``None``, all
        chains are included.
    max_chains
        Upper bound how many chains are included within each subplot/facet. Avoids
        overplotting. If ``None``, all chains contained in the ``results`` input are
        plotted. Always starts chain selection from the lowest chain index upwards. For
        selecting specific chains use the argument ``chain_indices``.
    title
        Plot title.
    title_spacing
        Determines the margin/whitespace between the plot title (set with
        ``fig.suptitle()``) and the first row of subplots/facets. Passed to the ``top``
        argument of ``fig.subplots_adjust()``.
    xlabel
        Label of the x-axis.
    style
        Passed to the ``style`` argument of ``sns.set_theme()``. Valid options are
        ``"darkgrid"``, ``"whitegrid"``, ``"dark"``, ``"white"``, and ``"ticks"``.
    color_palette
        Passed to the palette argument of ``sns.relplot()``. String values must be
        valid inputs of ``sns.color_palette()`` such as a seaborn color palette or a
        matplotlib colormap. Custom colors can be set with a list of color strings or a
        dictionary with the chain indices as keys and color strings as values. The
        number of color strings must coincide with the number of plotted chains. If
        ``None``, the default ``tab10`` matplotlib colormap is chosen.
    ncol
        Number of subplots/facets within each row of the grid.
    height
        Height in inches of each subplot/facet within the grid.
    aspect_ratio
        Ratio of width / height of each subplot/facet within the grid, i.e. ``width =
        aspect_ratio * height``.
    save_path
        File path where the plot is saved.
    include_warmup
        Include the warmup samples in the trace plot.
    **kwargs
        Further keyword arguments passed to the seaborn ``relplot()`` function.

    Returns
    -------
    A seaborn ``FacetGrid``.
    """

    # NOTE: Docstring duplications
    # The entries `results` to `max_chains` are shared with the `summary()`
    # and all user plotting functions.
    # The entries `title` to `save_path` are shared with `plot_density()`, `plot_cor()`
    # and partially with `plot_param()`.

    sns.set_theme(style=style)
    plot_df = _setup_plot_df(
        results, params, param_indices, chain_indices, max_chains, include_warmup
    )

    g = sns.relplot(
        data=plot_df,
        kind="line",
        x="iteration",
        y="value",
        hue="chain_index",
        col="param_label",
        col_wrap=_set_plot_cols(plot_df, ncol),
        facet_kws=dict(sharex=True, sharey=False),
        palette=color_palette,
        height=height,
        aspect=aspect_ratio,
        **kwargs,
    )

    g = _set_aesthetics(g, title, title_spacing, xlabel, ylabel="")

    save_figure(g, save_path)
    return g


def plot_density(
    results: SamplingResults,
    params: str | list[str] | None = None,
    param_indices: int | Sequence[int] | None = None,
    chain_indices: int | Sequence[int] | None = None,
    max_chains: int | None = 5,
    title: str | None = None,
    title_spacing: float = 0.85,
    xlabel: str = "Value",
    style: str = "whitegrid",
    color_palette: str | list[str] | dict[int, str] | None = None,
    ncol: int = 3,
    height: int = 3,
    aspect_ratio: int = 1,
    save_path: str | None = None,
    **kwargs,
) -> sns.FacetGrid:
    """
    Visualizes posterior distributions with a density plot.

    Parameters
    ----------
    results
        Result object of the sampling process. Must have a method
        ``get_posterior_samples()`` which extracts all samples from the posterior
        distribution.
    params
        Names of the model parameters that are contained in the plot. Must coincide
        with the dictionary keys of the ``Position`` with the posterior samples. If
        ``None``, all parameters are included.
    param_indices
        Indices of each model parameter that are contained in the plot. Selects e.g.
        ``beta[0]`` out of a ``beta`` parameter vector. A single index can be specified
        as an integer or a sequence containing one integer. If ``None``, all
        subparameters are included.
    chain_indices
        Indices of chains for each model subparameter that are contained in the plot.
        Selects e.g. chain 0 and chain 2 out of multiple chains. A single index can be
        specified as an integer or a sequence containing one integer. If ``None``, all
        chains are included.
    max_chains
        Upper bound how many chains are included within each subplot/facet. Avoids
        overplotting. If ``None``, all chains contained in the ``results`` input are
        plotted. Always starts chain selection from the lowest chain index upwards. For
        selecting specific chains use the argument ``chain_indices``.
    title
        Plot title.
    title_spacing
        Determines the margin/whitespace between the plot title (set with
        ``fig.suptitle()``) and the first row of subplots/facets. Passed to the ``top``
        argument of ``fig.subplots_adjust()``.
    xlabel
        Label of the x-axis.
    style
        Passed to the ``style`` argument of ``sns.set_theme()``. Valid options are
        ``darkgrid``, ``whitegrid``, ``dark``, ``white``, and ``ticks``.
    color_palette
        Passed to the palette argument of ``sns.displot()``. String values must be
        valid inputs of ``sns.color_palette()`` such as a seaborn color palette or a
        matplotlib colormap. Custom colors can be set with a list of color strings or a
        dictionary with the chain indices as keys and color strings as values. The
        number of color strings must coincide with the number of plotted chains. If
        ``None``, the default ``tab10`` matplotlib colormap is chosen.
    ncol
        Number of subplots/facets within each row of the grid.
    height
        Height in inches of each subplot/facet within the grid.
    aspect_ratio
        Ratio of width / height of each subplot/facet within the grid, i.e. ``width =
        aspect_ratio * height``.
    save_path
        File path where the plot is saved.
    **kwargs
        Further keyword arguments passed to the seaborn ``displot()`` function.

    Returns
    -------
    A seaborn ``FacetGrid``.
    """

    # NOTE: Docstring duplications
    # The entries `results` to `max_chains` are shared with the `summary()`
    # and all user plotting functions.
    # The entries `title` to `save_path` are shared with `plot_trace()`, `plot_cor()`
    # and partially with `plot_param()`.

    sns.set_theme(style=style)
    plot_df = _setup_plot_df(results, params, param_indices, chain_indices, max_chains)

    g = sns.displot(
        data=plot_df,
        kind="kde",
        x="value",
        y=None,
        hue="chain_index",
        col="param_label",
        col_wrap=_set_plot_cols(plot_df, ncol),
        facet_kws=dict(sharex=False, sharey=False),
        palette=color_palette,
        height=height,
        aspect=aspect_ratio,
        **kwargs,
    )

    g = _set_aesthetics(g, title, title_spacing, xlabel, ylabel="")

    save_figure(g, save_path)
    return g


def _compute_max_lags(
    plot_df: pd.DataFrame,
    max_lags: int | None,
) -> int:
    """
    Determines number time lags that are shown on the x-axis of the autocorrelation
    plot.
    """

    num_iterations = plot_df["iteration"].max().item()
    max_lags = np.min([num_iterations, 30]) if max_lags is None else max_lags
    return max_lags


def plot_cor(
    results: SamplingResults,
    params: str | list[str] | None = None,
    param_indices: int | Sequence[int] | None = None,
    chain_indices: int | Sequence[int] | None = None,
    max_chains: int | None = 5,
    max_lags: int | None = None,
    title: str | None = None,
    title_spacing: float = 0.85,
    xlabel: str = "Lag",
    style: str = "whitegrid",
    color_palette: str | list[str] | dict[int, str] | None = None,
    ncol: int = 3,
    height: int = 3,
    aspect_ratio: int = 1,
    save_path: str | None = None,
    **kwargs,
) -> sns.FacetGrid:
    """
    Visualizes autocorrelations of posterior samples.

    Parameters
    ----------
    results
        Result object of the sampling process. Must have a method
        ``get_posterior_samples()`` which extracts all samples from the posterior
        distribution.
    params
        Names of the model parameters that are contained in the plot. Must coincide
        with the dictionary keys of the ``Position`` with the posterior samples. If
        ``None``, all parameters are included.
    param_indices
        Indices of each model parameter that are contained in the plot. Selects e.g.
        ``beta[0]`` out of a ``beta`` parameter vector. A single index can be specified
        as an integer or a sequence containing one integer. If ``None``, all
        subparameters are included.
    chain_indices
        Indices of chains for each model subparameter that are contained in the plot.
        Selects e.g. chain 0 and chain 2 out of multiple chains. A single index can be
        specified as an integer or a sequence containing one integer. If ``None``, all
        chains are included.
    max_chains
        Upper bound how many chains are included within each subplot/facet. Avoids
        overplotting. If ``None``, all chains contained in the ``results`` input are
        plotted. Always starts chain selection from the lowest chain index upwards. For
        selecting specific chains use the argument ``chain_indices``.
    max_lags
        Maximum number of time lags shown on the x-axis of the autocorrelation plot. If
        ``None``, the minimum of the chain lengths and 30 is chosen.
    title
        Plot title.
    title_spacing
        Determines the margin/whitespace between the plot title (set with
        ``fig.suptitle()``) and the first row of subplots/facets. Passed to the ``top``
        argument of ``fig.subplots_adjust()``.
    xlabel
        Label of the x-axis.
    style
        Passed to the ``style`` argument of ``sns.set_theme()``. Valid options are
        ``darkgrid``, ``whitegrid``, ``dark``, ``white``, and ``ticks``.
    color_palette
        Passed to the palette argument of ``sns.FacetGrid()``. String values must be
        valid inputs of ``sns.color_palette()`` such as a seaborn color palette or a
        matplotlib colormap. Custom colors can be set with a list of color strings or a
        dictionary with the chain indices as keys and color strings as values. The
        number of color strings must coincide with the number of plotted chains. If
        ``None``, the default ``tab10`` matplotlib colormap is chosen.
    ncol
        Number of subplots/facets within each row of the grid.
    height
        Height in inches of each subplot/facet within the grid.
    aspect_ratio
        Ratio of width / height of each subplot/facet within the grid, i.e. ``width =
        aspect_ratio * height``.
    save_path
        File path where the plot is saved.
    **kwargs
        Further keyword arguments passed to the seaborn ``FacetGrid()`` function.


    Returns
    -------
    A seaborn ``FacetGrid``.
    """

    # NOTE: Docstring duplications
    # The entries `results` to `max_chains` are shared with the `summary()`
    # and all user plotting functions.
    # The entries `title` to `save_path` are shared with `plot_trace()`,
    # `plot_density()` and partially with `plot_param()`.
    # The entry `max_lags` is shared with `plot_param()`.

    sns.set_theme(style=style)
    plot_df = _setup_plot_df(results, params, param_indices, chain_indices, max_chains)
    max_lags = _compute_max_lags(plot_df, max_lags)

    def do_acor_plot(x, maxlags, **kwargs):
        x = np.asarray(x)
        acor = arviz.autocorr(x)[..., 0:maxlags]
        return sns.lineplot(x=range(maxlags), y=acor, **kwargs)

    g = (
        sns.FacetGrid(
            data=plot_df,
            hue="chain_index",
            col="param_label",
            col_wrap=_set_plot_cols(plot_df, ncol),
            palette=color_palette,
            height=height,
            aspect=aspect_ratio,
            **kwargs,
        )
        .map(
            do_acor_plot,
            "value",
            maxlags=max_lags,
        )
        .set(
            xlim=(0, max_lags),
            ylim=(-0.2, 1.1),
        )
        .add_legend()
    )

    g = _set_aesthetics(g, title, title_spacing, xlabel, ylabel="Autocorrelation")

    save_figure(g, save_path)
    return g


def _raise_multi_param_error(plot_df: pd.DataFrame, param: str) -> None:
    """
    :func:`.plot_param` function can only display all three diagnostic plots for a
    single subparameter. Throws an informative error otherwise.
    """

    if plot_df["param_label"].nunique() > 1:
        raise ValueError(
            f"{param} has more than one index. "
            "Please specify a single `param_index` for plotting."
        )


def _set_colors(plot_df: pd.DataFrame, color_list: list[str] | None) -> list[str]:
    """Determines colors of different chains in each plot."""

    num_chains = plot_df["chain_index"].nunique()

    if color_list is None:
        # default matplotlib and seaborn colors with 10 elements
        color_list = sns.color_palette()

        # make default color list sufficiently long
        if num_chains > 10:
            color_list = color_list * (num_chains // 10 + 1)

        color_list = color_list[:num_chains]  # type: ignore

    return color_list


def _setup_grid(
    figure_size: tuple[int | float, int | float]
) -> tuple[plt.Figure, Any, Any, Any]:
    """
    Initializes plotting grid with one large subplot for the trace plot and two smaller
    subplots for the density and autocorrelation plot.
    """

    fig = plt.figure(figsize=figure_size)
    ax1 = plt.subplot2grid(shape=(2, 2), loc=(0, 0), colspan=2)
    ax2 = plt.subplot2grid(shape=(2, 2), loc=(1, 0))
    ax3 = plt.subplot2grid(shape=(2, 2), loc=(1, 1))
    return fig, ax1, ax2, ax3


def _add_lineplot(plot_df: pd.DataFrame, ax: Any, color_list: list[str]) -> None:
    """Adds trace plot to plotting grid."""

    sns.lineplot(
        data=plot_df,
        x="iteration",
        y="value",
        hue="chain_index",
        palette=color_list,
        ax=ax,
        legend="full",
    ).set(xlabel="Iteration", ylabel="")


def _add_kdeplot(plot_df: pd.DataFrame, ax: Any, color_list: list[str]) -> None:
    """Adds density plot to plotting grid."""

    sns.kdeplot(
        data=plot_df,
        x="value",
        hue="chain_index",
        palette=color_list,
        ax=ax,
        legend=False,
    ).set(xlabel="Value", ylabel="")


def _add_corplot(
    plot_df: pd.DataFrame, ax: Any, max_lags: int | None, color_list: list[str]
) -> None:
    """Adds correlation plot to plotting grid."""

    max_lags = _compute_max_lags(plot_df, max_lags)

    for chain_index, col in zip(plot_df["chain_index"].unique(), color_list):
        x = np.asarray(plot_df.loc[plot_df["chain_index"] == chain_index]["value"])
        acor = arviz.autocorr(x)[0:max_lags]

        sns.lineplot(
            x=range(max_lags),
            y=acor,
            marker="",
            linestyle="-",
            color=col,
            ax=ax,
        )

        ax.set(xlim=(0, max_lags), ylim=(-0.2, 1.1), xlabel="Lag", ylabel="")


def _get_title(plot_df: pd.DataFrame, title: str | None) -> str:
    """Sets either a custom or the default plot title."""

    default_title = f"Diagnostic plots for '{plot_df['param'][0]}'"
    return title if title is not None else default_title


def plot_param(
    results: SamplingResults,
    param: str,
    param_index: int | None = None,
    chain_indices: int | Sequence[int] | None = None,
    max_chains: int | None = 5,
    max_lags: int | None = None,
    title: str | None = None,
    title_spacing: float = 0.9,
    style: str = "whitegrid",
    color_list: list[str] | None = None,
    figure_size: tuple[int | float, int | float] = (9, 6),
    # default values chosen for default figure size of (9, 6)
    legend_position: tuple[float, float] = (1.2, 0.4),
    save_path: str | None = None,
) -> None:
    """
    Visualizes trace plot, density plot and autocorrelation plot of a single
    subparameter.

    Parameters
    ----------
    results
        Result object of the sampling process. Must have a method
        ``get_posterior_samples()`` which extracts all samples from the posterior
        distribution.
    param
        Name of a single model parameter that is contained in the plot. Must coincide
        with one dictionary key of the ``Position`` with the posterior samples.
    param_index
        A single index of the selected model parameter that is contained in the plot.
        Selects e.g. ``beta[0]`` out of a ``beta`` parameter vector. Can be specified
        as an integer or as a sequence containing one integer. If ``None``, the
        parameter is assumed to have only a single index.
    chain_indices
        Indices of chains for each model subparameter that are contained in the plot.
        Selects e.g. chain 0 and chain 2 out of multiple chains. A single index can be
        specified as an integer or a sequence containing one integer. If ``None``, all
        chains are included.
    max_chains
        Upper bound how many chains are included within each subplot/facet. Avoids
        overplotting. If ``None``, all chains contained in the ``results`` input are
        plotted. Always starts chain selection from the lowest chain index upwards. For
        selecting specific chains use the argument ``chain_indices``.
    max_lags
        Maximum number of time lags shown on the x-axis of the autocorrelation plot. If
        ``None``, the minimum of the chain lengths and 30 is chosen.
    title
        Plot title.
    title_spacing
        Determines the margin/whitespace between the plot title (set with
        ``fig.suptitle()``) and the first row of subplots/facets. Passed to the ``top``
        argument of ``fig.subplots_adjust()``.
    style
        Passed to the ``style`` argument of ``sns.set_theme()``. Valid options are
        ``darkgrid``, ``whitegrid``, ``dark``, ``white``, and ``ticks``.
    color_list
        Determines the chain colors for all three subplots. Custom colors can be passed
        with a list of color strings. The length of the list must match the number of
        chains. If ``None``, the default ``tab10`` matplotlib colormap is chosen.
    figure_size
        Size of the entire plot grid. Passed to the ``figsize`` argument of
        ``plt.figure()``. When changing the figure size consider changing the
        ``legend_position`` as well. Generally, a ratio of 3
    legend_position
        Determines the color legend position. Coordinates are relative to the upper
        panel within the plot grid. The first coordinate specifies the horizontal, the
        second coordinate the vertical position. Might require an adjustment when
        changing the ``figure_size`` values or the number of chains.
    save_path
        File path where the plot is saved.
    """

    # NOTE: Docstring duplications
    # The entries `results`, `chain_indices` and `max_chains` are shared with the
    # `summary()` and all user plotting functions.
    # The entries `title`, `title_spacing`, `style` and `save_path` are shared with
    # `plot_trace()`, `plot_density()` and `plot_cor()`.
    # The entry `max_lags` is shared with `plot_cor()`.

    sns.set_theme(style=style)
    plot_df = _setup_plot_df(results, param, param_index, chain_indices, max_chains)
    _raise_multi_param_error(plot_df, param)
    color_list = _set_colors(plot_df, color_list)

    fig, ax1, ax2, ax3 = _setup_grid(figure_size)
    _add_lineplot(plot_df, ax1, color_list)
    _add_kdeplot(plot_df, ax2, color_list)
    _add_corplot(plot_df, ax3, max_lags, color_list)
    ax1.legend(title="Chain", bbox_to_anchor=legend_position, frameon=False)

    fig.tight_layout()
    fig.suptitle(_get_title(plot_df, title))
    fig.subplots_adjust(top=title_spacing)

    save_figure(save_path=save_path)


def plot_scatter(
    results: SamplingResults,
    params: list[str],
    param_indices: tuple[int, int],
    chain_indices: int | Sequence[int] | None = None,
    max_chains: int | None = 5,
    alpha: float = 0.2,
    title: str | None = None,
    title_spacing: float = 0.9,
    style: str = "whitegrid",
    color_list: list[str] | None = None,
    figure_size: tuple[int | float, int | float] = (9, 6),
    legend_position: tuple[float, float] | str = "best",
    save_path: str | None = None,
    include_warmup: bool = False,
):
    """
    Produces a scatterplot of two parameters.

    Parameters
    ----------
    results
        Result object of the sampling process. Must have a method
        ``get_posterior_samples()`` which extracts all samples from the posterior
        distribution.
    params
        Names of the model parameters that are contained in the plot. Must coincide with
        the dictionary keys of the ``Position`` with the posterior samples.
    param_indices
        Indices of each model parameter that are contained in the plot. Selects e.g.
        ``beta[0]`` out of a ``beta`` parameter vector. If only one string is supplied
        as the value of ``params``, ``param_indices`` must contain two indices. If a
        sequence of two strings is supplied to ``params``, you can supply either a
        single integer or a tuple of two integers. A single integer will be used as the
        index for *both* parameters. If you use a tuple of two integers, the first
        element will be used as the index for the first parameter, and the second
        element will be used as the index for the second parameter.
    chain_indices
        Indices of chains for each model subparameter that are contained in the plot.
        Selects e.g. chain 0 and chain 2 out of multiple chains. A single index can be
        specified as an integer or a sequence containing one integer. If ``None``, all
        chains are included.
    max_chains
        Upper bound how many chains are included within each subplot/facet. Avoids
        overplotting. If ``None``, all chains contained in the ``results`` input are
        plotted. Always starts chain selection from the lowest chain index upwards. For
        selecting specific chains use the argument ``chain_indices``.
    alpha
        Amount of transparency; a float between 0 and 1.
    title
        Plot title.
    title_spacing
        Determines the margin/whitespace between the plot title (set with
        ``fig.suptitle()``) and the first row of subplots/facets. Passed to the ``top``
        argument of ``fig.subplots_adjust()``.
    style
        Passed to the ``style`` argument of ``sns.set_theme()``. Valid options are
        ``darkgrid``, ``whitegrid``, ``dark``, ``white``, and ``ticks``.
    color_list
        Determines the chain colors for all three subplots. Custom colors can be passed
        with a list of color strings. The length of the list must match the number of
        chains. If ``None``, the default ``tab10`` matplotlib colormap is chosen.
    figure_size
        Size of the entire plot grid. Passed to the ``figsize`` argument of
        ``plt.figure()``. When changing the figure size consider changing the
        ``legend_position`` as well. Generally, a ratio of 3
    legend_position
        Determines the color legend position. Coordinates are relative to the upper
        panel within the plot grid. The first coordinate specifies the horizontal, the
        second coordinate the vertical position. Might require an adjustment when
        changing the ``figure_size`` values or the number of chains.
    save_path
        File path where the plot is saved.
    """
    # NOTE: Docstring duplications
    # Multiple arguments in this docstring are shared with other plotting functions.

    sns.set_theme(style=style)
    plot_df = _setup_scatterplot_df(
        results, params, param_indices, chain_indices, max_chains, include_warmup
    )

    labels = plot_df.param_label.unique()
    if len(labels) != 2:
        raise ValueError(
            "'plot_scatter' can only plot exactly two parameters. Use 'plot_pairs'"
            " instead to plot more."
        )

    plot_df = (
        plot_df.drop(["param_index", "param"], axis=1)
        .pivot(
            index=["chain_index", "iteration"], columns="param_label", values="value"
        )
        .reset_index()
        .drop(["iteration"], axis=1)
    )

    color_list = _set_colors(plot_df, color_list)
    fig, axis = plt.subplots(1, 1, figsize=figure_size)
    sns.scatterplot(
        data=plot_df,
        x=labels[0],
        y=labels[1],
        alpha=alpha,
        hue="chain_index",
        palette=color_list,
        ax=axis,
    )
    if title is not None:
        fig.suptitle()
        fig.subplots_adjust(top=title_spacing)

    axis.legend(title="Chain", loc=legend_position, frameon=False)

    save_figure(save_path=save_path)


def plot_pairs(
    results: SamplingResults,
    params: str | list[str] | None = None,
    param_indices: int | Sequence[int] | None = None,
    chain_indices: int | Sequence[int] | None = None,
    max_chains: int | None = 5,
    alpha: float = 0.2,
    title: str | None = None,
    title_spacing: float = 0.9,
    style: str = "whitegrid",
    diag_kind: str = "kde",
    color_palette: str | list[str] | dict[int, str] | None = None,
    height: int = 3,
    aspect_ratio: int = 1,
    save_path: str | None = None,
    include_warmup: bool = False,
):
    """
    Produces a pairplot panel.

    Parameters
    ----------
    results
        Result object of the sampling process. Must have a method
        ``get_posterior_samples()`` which extracts all samples from the posterior
        distribution.
    params
        Names of the model parameters that are contained in the plot. Must coincide with
        the dictionary keys of the ``Position`` with the posterior samples. If ``None``,
        all parameters are included.
    param_indices
        Indices of each model parameter that are contained in the plot. Selects e.g.
        ``beta[0]`` out of a ``beta`` parameter vector. A single index can be specified
        as an integer or a sequence containing one integer. If ``None``, all
        subparameters are included.
    chain_indices
        Indices of chains for each model subparameter that are contained in the plot.
        Selects e.g. chain 0 and chain 2 out of multiple chains. A single index can be
        specified as an integer or a sequence containing one integer. If ``None``, all
        chains are included.
    max_chains
        Upper bound how many chains are included within each subplot/facet. Avoids
        overplotting. If ``None``, all chains contained in the ``results`` input are
        plotted. Always starts chain selection from the lowest chain index upwards. For
        selecting specific chains use the argument ``chain_indices``.
    alpha
        Amount of transparency; a float between 0 and 1.
    title
        Plot title.
    title_spacing
        Determines the margin/whitespace between the plot title (set with
        ``fig.suptitle()``) and the first row of subplots/facets. Passed to the ``top``
        argument of ``fig.subplots_adjust()``.
    style
        Passed to the ``style`` argument of ``sns.set_theme()``. Valid options are
        ``darkgrid``, ``whitegrid``, ``dark``, ``white``, and ``ticks``.
    diag_kind
        Kind of plot for the diagonal subplots. Can be 'kde' (default) for kernel
        density estimates or 'hist' for histograms.
    color_palette
        Passed to the palette argument of ``sns.pairplot()``. String values must be
        valid inputs of ``sns.color_palette()`` such as a seaborn color palette or a
        matplotlib colormap. Custom colors can be set with a list of color strings or a
        dictionary with the chain indices as keys and color strings as values. The
        number of color strings must coincide with the number of plotted chains. If
        ``None``, the default ``tab10`` matplotlib colormap is chosen.
    height
        Height in inches of each subplot/facet within the grid.
    aspect_ratio
        Ratio of width / height of each subplot/facet within the grid, i.e. ``width =
        aspect_ratio * height``.
    legend_position
        Determines the color legend position. Coordinates are relative to the upper
        panel within the plot grid. The first coordinate specifies the horizontal, the
        second coordinate the vertical position. Might require an adjustment when
        changing the ``figure_size`` values or the number of chains.
    save_path
        File path where the plot is saved.
    include_warmup
        Include the warmup samples in the trace plot.
    """
    # NOTE: Docstring duplications
    # Multiple arguments in this docstring are shared with other plotting functions.

    sns.set_theme(style=style)
    plot_df = _setup_plot_df(
        results, params, param_indices, chain_indices, max_chains, include_warmup
    )

    plot_df = (
        plot_df.drop(["param_index", "param"], axis=1)
        .pivot(
            index=["chain_index", "iteration"], columns="param_label", values="value"
        )
        .reset_index()
        .drop(["iteration"], axis=1)
    )

    g = sns.pairplot(
        data=plot_df,
        hue="chain_index",
        plot_kws={"alpha": alpha},
        diag_kind=diag_kind,
        height=height,
        aspect=aspect_ratio,
        palette=color_palette,
    )

    if title is not None:
        g.fig.suptitle(title)
        g.fig.subplots_adjust(top=title_spacing)

    save_figure(save_path=save_path)
